# -*- coding: utf-8 -*-
"""SQLAlchemy-powered model definitions for profiles."""

from json import dumps

import colander
from sqlalchemy import Column, ForeignKey, String, Text, PickleType
from sqlalchemy.orm import relationship
from lxml import etree
from webhelpers2.html import HTML, literal

from ..lib.i18n import _, record_format_i18n
from ..lib.i18n import schema_i18n_labels, defaults_i18n_labels
from ..lib.i18n import view_i18n_labels, edit_i18n_labels
from ..lib.utils import make_id
from ..lib.xml import i18n_xml_text, db2xml_i18n_labels
from . import DBDeclarativeClass, ID_LEN, LABEL_LEN, DESCRIPTION_LEN
from .dbbase import DBBaseClass


# =============================================================================
class DBProfile(DBDeclarativeClass, DBBaseClass):
    """SQLAlchemy-powered profile class."""

    suffix = 'ciopfl'
    _settings_tabs = (_('Information'), _('Permissions'))

    __tablename__ = 'profiles'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    profile_id = Column(String(ID_LEN), primary_key=True)
    i18n_label = Column(Text(), nullable=False)
    i18n_description = Column(PickleType(1))

    principals = relationship('DBProfilePrincipal', cascade='all, delete')

    # -------------------------------------------------------------------------
    @classmethod
    def xml2db(cls, dbsession, profile_elt, error_if_exists=True, kwargs=None):
        """Load a profile from a XML element.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        :type  profile_elt: lxml.etree.Element
        :param profile_elt:
            Profile XML element.
        :param bool error_if_exists: (default=True)
            It returns an error if profile already exists.
        :param dict kwargs: (optional)
            Dictionary of keyword arguments.
        :rtype: :class:`pyramid.i18n.TranslationString` or ``None``
        :return:
            Error message or ``None``.
        """
        # pylint: disable = unused-argument
        # Check if already exists
        profile_id = make_id(profile_elt.get('id'), 'token', ID_LEN)
        dbprofile = dbsession.query(cls).filter_by(
            profile_id=profile_id).first()
        if dbprofile is not None:
            if error_if_exists:
                return _('Profile "${p}" already exists.', {'p': profile_id})
            return None

        # Create profile
        record = cls.record_from_xml(profile_id, profile_elt)
        error = cls.record_format(record)
        if error:
            return error
        dbprofile = cls(**record)
        dbsession.add(dbprofile)

        # Add principals
        dbsession.flush()
        for elt in profile_elt.findall('principals/principal'):
            dbprofile.principals.append(
                DBProfilePrincipal(principal=elt.text))

        return None

    # -------------------------------------------------------------------------
    @classmethod
    def record_from_xml(cls, profile_id, profile_elt):
        """Convert an profile XML element into a dictionary.

        :param str profile_id:
            Profile ID.
        :type  profile_elt: lxml.etree.Element
        :param profile_elt:
            Profile XML element.
        :rtype: dict
        """
        return {
            'profile_id': profile_id,
            'i18n_label': dumps(
                i18n_xml_text(profile_elt, 'label'), ensure_ascii=False),
            'i18n_description': i18n_xml_text(profile_elt, 'description')}

    # -------------------------------------------------------------------------
    @classmethod
    def record_format(cls, record):
        """Check and possibly correct a record before inserting it in the
        database.

        :param dict record:
            Dictionary of values to check.
        :rtype: ``None`` or :class:`pyramid.i18n.TranslationString`
        :return:
            ``None`` or error message.
        """
        for k in [i for i in record if record[i] is None]:
            del record[k]

        # Profile ID
        if not record.get('profile_id'):
            return _('Profile without ID.')
        record['profile_id'] = make_id(record['profile_id'], 'xmlid', ID_LEN)

        # Labels and descriptions
        if not record_format_i18n(record):
            return _('Profile without label.')
        if not record.get('i18n_description'):
            record['i18n_description'] = {}
        return None

    # -------------------------------------------------------------------------
    def db2xml(self, dbsession=None):
        """Serialize an profile to a XML representation.

        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession: (optional)
            SQLAlchemy session.
        :rtype: lxml.etree.Element
        """
        # pylint: disable = unused-argument
        profile_elt = etree.Element('profile')
        profile_elt.set('id', self.profile_id)

        # Labels and descriptions
        db2xml_i18n_labels(self, profile_elt, 3)

        # Principals
        if self.principals:
            elt = etree.SubElement(profile_elt, 'principals')
            for dbprincipal in self.principals:
                etree.SubElement(elt, 'principal').text = dbprincipal.principal

        return profile_elt

    # -------------------------------------------------------------------------
    def tab4view(self, request, tab_index, form):
        """Generate the tab content of a profile.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param int index:
            Index of the tab.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        if tab_index == 0:
            return self._tab4view_information(request, form)
        if tab_index == 1:
            return self._tab4view_permissions(request)
        return ''

    # -------------------------------------------------------------------------
    def _tab4view_information(self, request, form):
        """Generate the information tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        html = form.grid_item(
            translate(_('Identifier:')), self.profile_id, clear=True)
        html += view_i18n_labels(request, form, self)
        return html

    # -------------------------------------------------------------------------
    def _tab4view_permissions(self, request):
        """Generate the permission tab.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: webhelpers2.html.literal
        """
        active_mode = 'modes' in request.registry
        translate = request.localizer.translate
        html = ''
        active = [k.principal for k in self.principals]
        for group in request.registry['principals']:
            if not active_mode and group[0] == 'mode':
                continue
            html += HTML.h3(translate(group[1]))
            for principal in group[2]:
                if '{0}.{1}'.format(group[0], principal[0]) in active:
                    html += HTML.div('✔ {0}'.format(translate(principal[1])))
                else:
                    html += HTML.div('☐ {0}'.format(translate(principal[1])))
        return html

    # -------------------------------------------------------------------------
    @classmethod
    def settings_schema(cls, request, dbprofile=None):
        """Return a Colander schema to edit profile.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  dbprofile: DBProfile
        :param dbprofile: (optional)
            Current profile SqlAlchemy object.
        :rtype: tuple
        :return:
            A tuple such as ``(schema, defaults)``.
        """
        # Schema
        schema = colander.SchemaNode(colander.Mapping())
        if dbprofile is None:
            schema.add(colander.SchemaNode(
                colander.String(), name='profile_id',
                validator=colander.All(
                    colander.Regex(r'^[a-z][a-z0-9_]+$'),
                    colander.Length(min=2, max=ID_LEN))))

        schema_i18n_labels(request, schema, LABEL_LEN, DESCRIPTION_LEN)

        for group in request.registry['principals']:
            for principal in group[2]:
                schema.add(colander.SchemaNode(
                    colander.Boolean(), name='pcpl:{0}.{1}'.format(
                        group[0], principal[0]), missing=False))

        # Defaults
        if dbprofile is None:
            defaults = {}
        else:
            defaults = defaults_i18n_labels(dbprofile)
            for dbprincipal in dbprofile.principals:
                defaults['pcpl:{0}'.format(dbprincipal.principal)] = True

        return schema, defaults

    # -------------------------------------------------------------------------
    @classmethod
    def tab4edit(cls, request, tab_index, form, dbprofile=None):
        """Generate the tab content of profile for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param int tab_index:
            Index of the tab.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  dbprofile: DBProfile
        :param dbprofile: (optional)
            Current profile SqlAlchemy object.
        :rtype: webhelpers2.html.literal
        """
        if tab_index == 0:
            return cls._tab4edit_information(request, form, dbprofile)
        if tab_index == 1:
            return cls._tab4edit_permissions(request, form)
        return ''

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_information(cls, request, form, dbprofile):
        """Generate the information tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :type  dbprofile: DBProfile
        :param dbprofile:
            Current profile SqlAlchemy object.
        :rtype: webhelpers2.html.literal
        """
        translate = request.localizer.translate
        if dbprofile is None:
            html = form.grid_text(
                'profile_id', translate(_('Identifier:')), required=True,
                maxlength=ID_LEN, clear=True)
        else:
            html = form.grid_item(
                translate(_('Identifier:')), dbprofile.profile_id, clear=True)
        html += edit_i18n_labels(request, form, LABEL_LEN, DESCRIPTION_LEN)
        return html

    # -------------------------------------------------------------------------
    @classmethod
    def _tab4edit_permissions(cls, request, form):
        """Generate the permission tab for edition.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :type  form: .lib.form.Form
        :param form:
            Current form object.
        :rtype: webhelpers2.html.literal
        """
        active_mode = 'modes' in request.registry
        translate = request.localizer.translate
        html = ''
        for group in request.registry['principals']:
            if not active_mode and group[0] == 'mode':
                continue
            html += HTML.h3(translate(group[1]))
            for principal in group[2]:
                html += HTML.div(
                    form.custom_checkbox('pcpl:{0}.{1}'.format(
                        group[0], principal[0])) +
                    literal(' <label for="pcpl{0}{1}">{2}</label>'.format(
                        group[0], principal[0], translate(principal[1]))))
        return html


# =============================================================================
class DBProfilePrincipal(DBDeclarativeClass):
    """SQLAlchemy-powered profile permission class (one-to-many)."""
    # pylint: disable = too-few-public-methods

    __tablename__ = 'profiles_principals'
    __table_args__ = {'mysql_engine': 'InnoDB'}

    profile_id = Column(
        String(ID_LEN), ForeignKey('profiles.profile_id', ondelete='CASCADE'),
        primary_key=True)
    principal = Column(String(ID_LEN), primary_key=True)



"""Unit, integration, and functional testing for LDAP module."""

from os.path import join, dirname

TEST1_LDAP_XML = join(dirname(__file__), 'Xml', 'test1.cioldap.xml')
TEST2_LDAP_XML = join(dirname(__file__), 'Xml', 'test2.cioldap.xml')



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``includes.cache`` classes and functions."""

from time import sleep, time
from unittest import TestCase

from pyramid.testing import DummyRequest

from . import ConfiguratorTestCase


# =============================================================================
class UIncludesCacheIncludeme(ConfiguratorTestCase):
    """Unit test class for :class:`includes.cache.includeme`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.cache.includeme]"""
        self.configurator.include('..includes.cache')
        self.assertIn('cache_user', self.configurator.registry)
        self.assertIn('cache_global', self.configurator.registry)


# =============================================================================
class UIncludesCacheCacheUser(TestCase):
    """Unit test class for :class:`includes.cache.CacheUser`."""

    # -------------------------------------------------------------------------
    def test_set(self):
        """[u:includes.cache.CacheUser.set]"""
        from ..includes.cache import CACHE_DEFAULT_TTL, CacheUser

        cache_user = CacheUser({
            'beaker.cache.expire': '43200',
            'beaker.cache.regions': 'default, short, long',
            'beaker.cache.short.expire': '1',
            'beaker.cache.long.expire': '1800'})
        request = DummyRequest()

        # Without namespace and region
        cache_user.set(request, 'foo', {'one': 1, 'two': 2})
        value = cache_user.get(request, 'foo')
        self.assertIsInstance(value, dict)
        self.assertIn('one', value)
        self.assertIsNone(cache_user.get(request, 'bar'))

        # With namespace and without region
        cache_user.set(request, 'baz', {'one': 1, 'two': 2}, 'namespace1')
        value = cache_user.get(request, 'baz', 'namespace1')
        self.assertIsInstance(value, dict)
        self.assertIn('two', value)
        self.assertIsNone(cache_user.get(request, 'baz'))
        self.assertIsNone(cache_user.get(request, 'baz', 'namespace2'))
        self.assertIsNone(cache_user.get(request, 'foo', 'namespace1'))

        # With namespace and unknown region
        cache_user.set(request, 'foo2', 2, 'namespace2', 'very_short')
        self.assertEqual(cache_user.get(request, 'foo2', 'namespace2'), 2)

        # With namespace and region
        cache_user.set(request, 'foo3', 3, 'namespace3', 'short')
        self.assertEqual(cache_user.get(request, 'foo3', 'namespace3'), 3)
        sleep(1.1)
        self.assertIsNone(cache_user.get(request, 'foo3', 'namespace3'))

        # With short expiration
        cache_user.set(request, 'foo', 'première', 'namespace4', expire=.1)
        self.assertEqual(
            cache_user.get(request, 'foo', 'namespace4'), 'première')
        sleep(.2)
        self.assertIsNone(cache_user.get(request, 'foo', 'namespace4'))

        # Replace cache
        cache_user.set(request, 'foo', 'three')
        self.assertEqual(cache_user.get(request, 'foo'), 'three')

        # Default region
        cache_user.set(request, 'foo', 'five', 'namespace5', 'default')
        self.assertIn('default', cache_user.regions)
        self.assertEqual(cache_user.regions['default'], CACHE_DEFAULT_TTL)

    # -------------------------------------------------------------------------
    def test_clear(self):
        """[u:includes.cache.CacheUser.clear]"""
        from ..includes.cache import CacheUser

        cache_user = CacheUser({
            'cache.expire': '43200',
            'cache.regions': 'long',
            'cache.long.expire': '1800'})
        request = DummyRequest()

        # Clear a key
        cache_user.set(request, 'pi', 3.14)
        cache_user.set(request, 'e', 2.71828)
        self.assertEqual(cache_user.get(request, 'pi'), 3.14)
        self.assertEqual(cache_user.get(request, 'e'), 2.71828)
        cache_user.clear(request, key='pi')
        self.assertIsNone(cache_user.get(request, 'pi'))
        self.assertEqual(cache_user.get(request, 'e'), 2.71828)

        # Clear an unknown key
        cache_user.clear(request, key='foo')
        self.assertEqual(cache_user.get(request, 'e'), 2.71828)

        # Clear a region
        cache_user.clear(request)
        self.assertIsNone(cache_user.get(request, 'e'))
        cache_user.clear(request)

    # -------------------------------------------------------------------------
    def test_purge(self):
        """[u:includes.cache.CacheUser.purge]"""
        from ..includes.cache import CacheUser

        cache_user = CacheUser({})
        request = DummyRequest()
        cache_user.set(request, 'foo', 'bar', expire=.1)
        sleep(.2)
        cache_user.purge(request)
        self.assertIsNone(cache_user.get(request, 'foo'))


# =============================================================================
class UIncludesCacheCacheUserAccess(TestCase):
    """Unit test class for :class:`includes.cache.cache_user_access`."""

    # -------------------------------------------------------------------------
    def test_without_uid(self):
        """[u:includes.cache.cache_user_access] without uid"""
        from ..includes.cache import cache_user_access

        # pylint: disable = too-few-public-methods
        class Item(object):
            """Class for items."""

        class ItemManager(object):
            """Class to manage items."""

            @cache_user_access('ns0')
            @classmethod
            def item_access(cls, request, item):
                """Method for testing."""
                # pylint: disable = unused-argument
                return ('writer',)

        item_manager = ItemManager()
        request = DummyRequest()
        item = Item()
        self.assertRaises(Exception, item_manager.item_access, request, item)

    # -------------------------------------------------------------------------
    def test_cache(self):
        """[u:includes.cache.cache_user_access] cache"""
        from ..includes.cache import CacheUser, cache_user_access

        # pylint: disable = too-few-public-methods
        class Item(object):
            """Class for items."""
            uid = 'itme01'

        class ItemManager(object):
            """Class to manage items."""

            @cache_user_access('ns0')
            @classmethod
            def item_access(cls, request, item):
                """Method for testing."""
                # pylint: disable = unused-argument
                return ('writer' if item is not None else None,)

        item_manager = ItemManager()
        request = DummyRequest()
        request.registry['cache_user'] = CacheUser({})
        item = Item()

        # With None
        result = item_manager.item_access(request, None)
        self.assertEqual(result, (None,))

        # First time, from method
        result = item_manager.item_access(request, item)
        self.assertEqual(result, ('writer',))

        # Second time, from cache
        result = item_manager.item_access(request, item)
        self.assertEqual(result, ('writer',))


# =============================================================================
class UIncludesCacheCacheGlobal(TestCase):
    """Unit test class for :class:`includes.cache.CacheGlobal`."""

    # -------------------------------------------------------------------------
    def test_get(self):
        """[u:includes.cache.CacheGlobal.(set|get)]"""
        from ..includes.cache import CACHE_DEFAULT_TTL, CacheGlobal

        cache_global = CacheGlobal({
            'beaker.cache.type': 'memory',
            'beaker.cache.regions': 'default, short, long',
            'beaker.cache.short.expire': '1',
            'beaker.cache.long.expire': '1800'})

        # Without namespace and region
        cache_global.set('foo', {'one': 1, 'two': 2})
        value = cache_global.get('foo')
        self.assertIsInstance(value, dict)
        self.assertIn('one', value)
        self.assertIsNone(cache_global.get('bar'))

        # With namespace and without region
        cache_global.set('baz', {'one': 1, 'two': 2}, 'namespace1')
        value = cache_global.get('baz', 'namespace1')
        self.assertIsInstance(value, dict)
        self.assertIn('two', value)
        self.assertIsNone(cache_global.get('baz'))
        self.assertIsNone(cache_global.get('baz', 'namespace2'))
        self.assertIsNone(cache_global.get('foo', 'namespace1'))

        # With namespace and unknown region
        cache_global.set('foo2', 2, 'namespace2', 'very_short')
        self.assertEqual(cache_global.get('foo2', 'namespace2'), 2)

        # With namespace and region
        cache_global.set('foo3', 3, 'namespace3', 'short')
        self.assertEqual(cache_global.get('foo3', 'namespace3'), 3)
        sleep(1.1)
        self.assertIsNone(cache_global.get('foo3', 'namespace3'))

        # With short expiration
        cache_global.set('foo', 'première', 'namespace4', expire=1)
        self.assertEqual(cache_global.get('foo', 'namespace4'), 'première')
        sleep(1.1)
        self.assertIsNone(cache_global.get('foo', 'namespace4'))

        # Replace cache
        cache_global.set('foo', 'three')
        self.assertEqual(cache_global.get('foo'), 'three')

        # Default region without global expiration
        cache_global.set('foo', 'five', 'namespace5', 'default')
        self.assertIn('default', cache_global.regions)
        self.assertIn('expire', cache_global.regions['default'])
        self.assertEqual(
            cache_global.regions['default']['expire'], CACHE_DEFAULT_TTL)

        # Default region with global expiration
        cache_global = CacheGlobal({
            'beaker.cache.type': 'memory',
            'beaker.cache.expire': '43200',
            'beaker.cache.regions': 'default'})
        cache_global.set('foo', 'six', 'namespace6', 'default')
        self.assertIn('default', cache_global.regions)
        self.assertIn('expire', cache_global.regions['default'])
        self.assertEqual(cache_global.regions['default']['expire'], 43200)

    # -------------------------------------------------------------------------
    def test_clear(self):
        """[u:includes.cache.CacheGlobal.clear]"""
        from ..includes.cache import CacheGlobal

        cache_global = CacheGlobal({
            'cache.expire': '43200',
            'cache.regions': 'long',
            'cache.long.expire': '1800'})

        # Clear a key
        cache_global.set('pi', 3.14)
        cache_global.set('e', 2.71828)
        self.assertEqual(cache_global.get('pi'), 3.14)
        self.assertEqual(cache_global.get('e'), 2.71828)
        cache_global.clear(key='pi')
        self.assertIsNone(cache_global.get('pi'))
        self.assertEqual(cache_global.get('e'), 2.71828)

        # Clear an unknown key
        cache_global.clear(key='foo')
        cache_global.clear(key='e', namespace='bar')
        self.assertEqual(cache_global.get('e'), 2.71828)

        # Clear a region
        cache_global.clear()
        self.assertIsNone(cache_global.get('e'))
        cache_global.clear()

    # -------------------------------------------------------------------------
    def test_initialize(self):
        """[u:includes.cache.CacheGlobal.initialize]"""
        from ..includes.cache import CREATED_KEY, CacheGlobal

        cache_global = CacheGlobal({
            'cache.expire': '43200',
            'cache.regions': 'long',
            'cache.long.expire': '1800'})
        cache_global.set('pi', 3.14)
        self.assertEqual(cache_global.get('pi'), 3.14)
        self.assertIsNone(cache_global.get(CREATED_KEY))
        now = time()

        cache_global.initialize()
        self.assertIsNone(cache_global.get('e'))
        created = cache_global.get(CREATED_KEY)
        self.assertIsInstance(created, float)
        self.assertGreater(created, now)
        cache_global.clear()


# =============================================================================
class UIncludesCacheCacheGlobalItem(TestCase):
    """Unit test class for :class:`includes.cache.cache_global_item`."""

    # -------------------------------------------------------------------------
    def test_without_uid(self):
        """[u:includes.cache.cache_global_item] without _items"""
        from ..includes.cache import cache_global_item

        class ItemManager(object):
            """Class to manage items."""

            @cache_global_item('ns0')
            @classmethod
            def item(cls, request, item_id):
                """Method for testing."""
                # pylint: disable = unused-argument
                return None

        item_manager = ItemManager()
        request = DummyRequest()
        self.assertRaises(Exception, item_manager.item, request, 'item01')

    # -------------------------------------------------------------------------
    def test_cache(self):
        """[u:includes.cache.cache_global_item] cache"""
        from ..includes.cache import CacheUser, CacheGlobal, cache_global_item

        item_ok = False

        # pylint: disable = too-few-public-methods
        class Item(object):
            """Class for items."""
            uid = 'itme01'
            created = time()

        class ItemManager(object):
            """Class to manage items."""
            _item1s = {}
            _item2s = {}

            @classmethod
            def item_access(cls, request, item):
                """Method for testing."""
                # pylint: disable = unused-argument
                return ('writer',)

            @cache_global_item('ns0')
            @classmethod
            def item1(cls, request, item_id):
                """Method for testing."""
                # pylint: disable = unused-argument
                return Item() if item_ok else None

            @cache_global_item('ns0', access_function=item_access)
            @classmethod
            def item2(cls, request, item_id):
                """Method for testing."""
                # pylint: disable = unused-argument
                return Item() if item_ok else None

        item_manager = ItemManager()
        request = DummyRequest()
        request.registry['cache_user'] = CacheUser({})
        request.registry['cache_global'] = CacheGlobal({})

        # With a bad item
        item = item_manager.item1(request, 'item01')
        self.assertIsNone(item)

        # With a good item, without access rights
        item_ok = True
        item = item_manager.item1(request, 'item01')
        self.assertIsInstance(item, Item)

        # With access rights
        item = item_manager.item2(request, 'item02')
        self.assertIsInstance(item, Item)

        # Second time, from cache
        item = item_manager.item2(request, 'item02')
        self.assertIsInstance(item, Item)


# =============================================================================
class UIncludesCacheCacheGlobalValue(TestCase):
    """Unit test class for :class:`includes.cache.cache_global_value`."""

    # -------------------------------------------------------------------------
    def test_without_uid(self):
        """[u:includes.cache.cache_global_value] without uid"""
        from ..includes.cache import cache_global_value

        class Item(object):
            """Class for items."""

            @cache_global_value('field1', 'ns0')
            @classmethod
            def get_field1(cls, request):
                """Method for testing."""
                # pylint: disable = unused-argument
                return .6

        item = Item()
        request = DummyRequest()
        self.assertRaises(Exception, item.get_field1, request)

    # -------------------------------------------------------------------------
    def test_cache(self):
        """[u:includes.cache.cache_global_value] cache"""
        from ..includes.cache import cache_global_value

        class Item(object):
            """Class for items."""
            uid = 'itme01'

            @cache_global_value('field1', 'ns0')
            @classmethod
            def get_field1(cls, request):
                """Method for testing."""
                # pylint: disable = unused-argument
                return .6

        item = Item()
        request = DummyRequest()

        # First time, from method
        result = item.get_field1(request)
        self.assertEqual(result, .6)

        # Second time, from cache
        result = item.get_field1(request)
        self.assertEqual(result, .6)


# =============================================================================
class UIncludesCacheCacheNamespace(TestCase):
    """Unit test class for :func:`includes.cache.cache_namespace`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:includes.cache.cache_namespace]"""
        from ..includes.cache import cache_namespace
        self.assertEqual(cache_namespace('ns0', 'foo'), 'ns0-foo')



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``models.dbprofile`` classes."""

from json import loads, dumps

from lxml import etree
from colander import SchemaNode

from pyramid.i18n import TranslationString

from . import DBUnitTestCase


# =============================================================================
class UModelsDBProfileDBProfile(DBUnitTestCase):
    """Unit test class for :class:`models.dbprofile.Profile`."""

    # -------------------------------------------------------------------------
    @classmethod
    def _make_one(cls):
        """Make an `DBProfile`` object."""
        from ..models.dbprofile import DBProfile

        return DBProfile(
            profile_id='user_creator',
            i18n_label=dumps({'en': 'Profile manager'}),
            i18n_description={'en': 'Permission to create profiles'})

    # -------------------------------------------------------------------------
    def test_xml2db(self):
        """[u:models.dbprofile.Profile.xml2db]"""
        from ..models.dbprofile import DBProfile

        profile_elt = etree.XML(
            '<profile id="profile_manager">'
            '  <label xml:lang="en">Profile manager</label>'
            '  <description xml:lang="en">'
            '    Permission to create, edit and view profiles.'
            '  </description>'
            '  <principals>'
            '    <principal>profile.creator</principal>'
            '  </principals>'
            '</profile>')
        dbsession = self.request.dbsession

        DBProfile.xml2db(dbsession, profile_elt)
        dbprofile = dbsession.query(DBProfile).filter_by(
            profile_id='profile_manager').first()
        self.assertIsInstance(dbprofile, DBProfile)
        self.assertIn('Profile manager', dbprofile.i18n_label)
        self.assertIsInstance(dbprofile.i18n_description, dict)
        self.assertIn('en', dbprofile.i18n_description)
        self.assertEqual(
            dbprofile.i18n_description['en'],
            'Permission to create, edit and view profiles.')
        self.assertIsNotNone(dbprofile.principals)
        self.assertEqual(len(dbprofile.principals), 1)

        error = DBProfile.xml2db(dbsession, profile_elt)
        self.assertIsInstance(error, TranslationString)
        error = DBProfile.xml2db(dbsession, profile_elt, error_if_exists=False)
        self.assertIsNone(error)

        profile_elt = etree.XML('<profile id="profile_editor"></profile>')
        error = DBProfile.xml2db(dbsession, profile_elt)
        self.assertIsInstance(error, TranslationString)

    # -------------------------------------------------------------------------
    def test_record_from_xml(self):
        """[u:models.dbprofile.DBProfile.record_from_xml]"""
        from ..models.dbprofile import DBProfile

        profile_elt = etree.XML(
            '<profile id="user_editor">'
            '  <label xml:lang="en">User account editor</label>'
            '  <label xml:lang="fr">Éditeur des comptes utilisateurs</label>'
            '  <description xml:lang="en">'
            '    These users can view and edit user accounts.'
            '  </description>'
            '  <principals>'
            '    <principal>user.editor</principal>'
            '  </principals>'
            '</profile>')
        record = DBProfile.record_from_xml('user_editor', profile_elt)
        self.assertIn('profile_id', record)
        self.assertEqual(record['profile_id'], 'user_editor')
        self.assertIn('i18n_label', record)
        labels = loads(record['i18n_label'])
        self.assertIn('fr', labels)
        self.assertEqual(labels['fr'], 'Éditeur des comptes utilisateurs')

    # -------------------------------------------------------------------------
    def test_record_format(self):
        """[u:models.dbprofile.DBProfile.record_format]"""
        from ..models.dbprofile import DBProfile

        record = {'profile_id': None}
        error = DBProfile.record_format(record)
        self.assertIsInstance(error, TranslationString)
        self.assertIn('without ID', error)

        record = {'profile_id': 'user_creator'}
        error = DBProfile.record_format(record)
        self.assertIsInstance(error, TranslationString)
        self.assertIn('without label', error)

        record = {
            'profile_id': 'user_creator', 'label_en': 'User account creator',
            'label_fr': 'Créateur de comptes utilisateurs',
            'description_en': 'These users can create user accounts'}
        error = DBProfile.record_format(record)
        self.assertIsNone(error)
        self.assertIn('i18n_label', record)
        labels = loads(record['i18n_label'])
        self.assertEqual(len(labels), 2)
        self.assertIn('fr', labels)
        self.assertEqual(labels['fr'], 'Créateur de comptes utilisateurs')
        self.assertIn('i18n_description', record)
        self.assertEqual(len(record['i18n_description']), 1)

    # -------------------------------------------------------------------------
    def test_db2xml(self):
        """[u:models.dbprofile.DBProfile.db2xml]"""
        from ..models.dbprofile import DBProfile, DBProfilePrincipal

        dbsession = self.request.dbsession
        dbprofile = DBProfile(
            profile_id='profile_manager',
            i18n_label=dumps({'en': 'Profile manager'}),
            i18n_description={'en': 'Permission to view, edit and create.'})
        dbprofile.principals.append(
            DBProfilePrincipal(principal='profile.creator'))

        profile_elt = dbprofile.db2xml(dbsession)
        self.assertEqual(profile_elt.get('id'), 'profile_manager')
        self.assertEqual(
            profile_elt.findtext('label'), 'Profile manager')
        self.assertEqual(
            profile_elt.findtext('description').strip(),
            'Permission to view, edit and create.')
        self.assertIsNotNone(profile_elt.find('principals'))
        self.assertEqual(
            profile_elt.findtext('principals/principal'), 'profile.creator')

    # -------------------------------------------------------------------------
    def test_tab4view(self):
        """[u:models.dbprofile.DBProfile.tab4view]"""
        from ..lib.form import Form
        from ..models.dbprofile import DBProfilePrincipal

        dbprofile = self._make_one()
        dbprofile.principals.append(
            DBProfilePrincipal(principal='user.creator'))
        self.configurator.add_route(
            'profile_view', '/profile/view/{profile_id}')
        principals = (
            ('user', 'User management', (
                ('editor', 'Edit or view any user', (
                    'user-edit', 'user-view')),
                ('creator', 'Create a new one or edit or view any user', (
                    'user-create', 'user-edit', 'user-view'))
            )),)
        self.request.registry['principals'] = principals

        html = dbprofile.tab4view(self.request, 0, Form(self.request))
        self.assertIn('Profile manager', html)

        html = dbprofile.tab4view(self.request, 1, Form(self.request))
        self.assertIn('Edit or view any user', html)

        self.assertEqual(
            dbprofile.tab4view(self.request, 2, Form(self.request)), '')

    # -------------------------------------------------------------------------
    def test_settings_schema(self):
        """[u:models.dbprofile.DBProfile.settings_schema]"""
        from ..models.dbprofile import DBProfile, DBProfilePrincipal

        self.request.registry.settings['languages'] = 'en, fr'
        principals = (
            ('user', 'User management', (
                ('editor', 'Edit or view any user', (
                    'user-edit', 'user-view')),
                ('creator', 'Create a new one or edit or view any user', (
                    'user-create', 'user-edit', 'user-view'))
            )),)
        self.request.registry['principals'] = principals

        # Creation
        schema, defaults = DBProfile.settings_schema(self.request)
        self.assertIsInstance(schema, SchemaNode)
        serialized = schema.serialize()
        self.assertIn('profile_id', serialized)
        self.assertIn('label_en', serialized)
        self.assertIn('label_fr', serialized)
        self.assertIn('description_en', serialized)
        self.assertIn('description_fr', serialized)
        self.assertIn('pcpl:user.editor', serialized)
        self.assertIsInstance(defaults, dict)
        self.assertFalse(defaults)

        # Update
        dbprofile = self._make_one()
        dbprofile.principals.append(
            DBProfilePrincipal(principal='user.creator'))
        schema, defaults = DBProfile.settings_schema(self.request, dbprofile)
        self.assertNotIn('profile_id', schema.serialize())
        self.assertIn('label_en', defaults)
        self.assertIn('description_en', defaults)
        self.assertIn('pcpl:user.creator', defaults)

    # -------------------------------------------------------------------------
    def test_tab4edit(self):
        """[u:models.dbprofile.DBProfile.tab4edit]"""
        from ..lib.form import Form
        from ..models.dbprofile import DBProfile

        self.request.registry.settings['languages'] = 'en, fr'
        principals = (
            ('user', 'User management', (
                ('editor', 'Edit or view any user', (
                    'user-edit', 'user-view')),
                ('creator', 'Create a new one or edit or view any user', (
                    'user-create', 'user-edit', 'user-view'))
            )),)
        self.request.registry['principals'] = principals
        dbprofile = self._make_one()

        # Information, creation
        html = DBProfile.tab4edit(self.request, 0, Form(self.request))
        self.assertIn('Identifier:', html)
        self.assertIn('profile_id', html)
        self.assertIn('label_en', html)
        self.assertIn('label_fr', html)
        self.assertIn('description_en', html)
        self.assertIn('description_fr', html)

        # Information, update
        html = DBProfile.tab4edit(
            self.request, 0, Form(self.request), dbprofile)
        self.assertNotIn('profile_id', html)

        # Principals
        html = DBProfile.tab4edit(self.request, 1, Form(self.request))
        self.assertIn('pcpl:user.editor', html)
        self.assertIn('for="pcplusereditor"', html)

        # Other
        html = DBProfile.tab4edit(self.request, 2, Form(self.request))
        self.assertEqual(html, '')


# =============================================================================
class IModelsDBProfile(DBUnitTestCase):
    """Integration test class for ``models.dbprofile``."""

    # -------------------------------------------------------------------------
    def test_delete_cascade_principal(self):
        """[i:models.dbprofile] delete, cascade principal"""
        from ..models.dbprofile import DBProfile, DBProfilePrincipal

        dbsession = self.request.dbsession
        profile_id = 'profile_manager'
        dbprofile = DBProfile(
            profile_id=profile_id,
            i18n_label=dumps({'en': 'Profile manager'}))
        dbprofile.principals.append(
            DBProfilePrincipal(principal='profile.creator'))
        dbsession.add(dbprofile)

        dbprofile = dbsession.query(DBProfile).filter_by(
            profile_id=profile_id).first()
        self.assertIsInstance(dbprofile, DBProfile)
        dbprincipal = dbsession.query(DBProfilePrincipal).filter_by(
            profile_id=profile_id).first()
        self.assertIsInstance(dbprincipal, DBProfilePrincipal)

        dbsession.delete(dbprofile)
        dbsession.flush()
        dbprofile = dbsession.query(DBProfile).filter_by(
            profile_id=profile_id).first()
        self.assertIsNone(dbprofile)
        dbprincipal = dbsession.query(DBProfilePrincipal).filter_by(
            profile_id=profile_id).first()
        self.assertIsNone(dbprincipal)



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``scripts.ciobackup`` function and class."""

from os import makedirs
from os.path import exists, join
from collections import namedtuple
from datetime import datetime
from shutil import rmtree
from unittest import TestCase

from transaction import manager

from . import DBUnitTestCase
# pylint: disable = unused-import
from ..includes.modules.models import DBModule  # noqa
from ..modules.cioskeleton.models.dbbone import DBBone  # noqa
# pylint: enable = unused-import


# =============================================================================
class UScriptsCIOBackupMain(TestCase):
    """Unit test class for :func:`scripts.ciobackup.main`."""

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Clear up temporary files."""
        from . import BACKUP_DIR

        if exists(BACKUP_DIR):
            rmtree(BACKUP_DIR)

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:scripts.ciobackup.main]"""
        from ..scripts.ciobackup import main
        from . import TEST_INI, BACKUP_DIR

        self.assertRaises(SystemExit, main)
        try:
            main([TEST_INI, BACKUP_DIR])
        except SystemExit:
            pass


# =============================================================================
class UScriptsBackup(DBUnitTestCase):
    """Unit test class for :class:`scripts.Backup`."""

    # -------------------------------------------------------------------------
    def test_argument_parser(self):
        """[u:scripts.Backup.argument_parser]"""
        from argparse import ArgumentParser
        from ..scripts.ciobackup import Backup

        parser = Backup.argument_parser()
        self.assertIsInstance(parser, ArgumentParser)
        self.assertEqual(parser.description, 'Backup database.')

    # -------------------------------------------------------------------------
    def test_run(self):
        """[u:scripts.Backup.run]"""
        # pylint: disable = too-many-locals
        from . import TEST_INI, TEST1_INI, TEST_DIR, BACKUP_DIR
        from ..relaxng import RELAXNG
        from ..scripts.ciobackup import Backup
        from ..models import get_dbsession_factory
        from ..models.populate import db2xml
        from ..models.dbsettings import DBSettings
        from ..models.dbuser import DBUser

        args = namedtuple('ArgumentParser', 'conf_uri options lang')(
            conf_uri=TEST1_INI, options=None, lang=None)
        self.assertEqual(Backup(args, db2xml, RELAXNG).run(BACKUP_DIR), 1)

        dbsession = self.request.dbsession
        args = namedtuple('ArgumentParser', 'conf_uri options lang')(
            conf_uri=TEST_INI, options=None, lang=None)
        self.assertEqual(Backup(
            args, db2xml, RELAXNG, dbsession_factory=get_dbsession_factory(
                dbsession.get_bind())).run(TEST_INI), 0)

        dbsession.add(DBSettings(
            key='populate', value=datetime.now().isoformat(' ').split('.')[0]))
        dbuser = DBUser(
            login='test1', first_name='Édith', last_name='AVULEUR',
            email='test1@chrysal.io')
        dbuser.set_password('test1pwd')
        dbsession.add(dbuser)
        manager.commit()
        makedirs(join(TEST_DIR, 'Attachments', 'Users'))

        args = namedtuple('ArgumentParser', 'conf_uri options lang')(
            conf_uri=TEST_INI, options=None, lang=None)
        self.assertEqual(Backup(
            args, db2xml, RELAXNG, dbsession_factory=get_dbsession_factory(
                dbsession.get_bind())).run(TEST_INI), 1)

        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang no_validation')(
                conf_uri=TEST_INI, options=None, lang=None,
                no_validation=False)
        self.assertEqual(Backup(
            args, db2xml, RELAXNG, dbsession_factory=get_dbsession_factory(
                dbsession.get_bind())).run(BACKUP_DIR), 0)
        self.assertTrue(exists(BACKUP_DIR))
        self.assertTrue(exists(join(BACKUP_DIR, 'testchrysalio.xml')))
        self.assertTrue(exists(join(BACKUP_DIR, 'Users')))

        if not exists(join(BACKUP_DIR, 'Foo')):
            makedirs(join(BACKUP_DIR, 'Foo'))
        if not exists(join(BACKUP_DIR, 'foo.xml')):
            with open(join(BACKUP_DIR, 'foo.xml'), 'w') as hdl:
                hdl.write('<chrysalio/>')
        self.assertEqual(Backup(
            args, db2xml, RELAXNG, dbsession_factory=get_dbsession_factory(
                dbsession.get_bind())).run(BACKUP_DIR), 0)
        self.assertFalse(exists(join(BACKUP_DIR, 'Foo')))
        self.assertTrue(exists(join(BACKUP_DIR, 'foo.xml')))



"""An example of a module for Chrysalio with dependencies, permissions, routes,
views and menu."""

from os.path import dirname, join

from pyramid.config import Configurator

from ...lib.i18n import _
from ...includes.modules.models import DBModule
from .. import Module
from .relaxng import RELAXNG_CIOSKELETON
from .security import PRINCIPALS_CIOSKELETON
from .menu import MENU_CIOSKELETON
from .modes import MODE_CIOSKELETON
from .models.populate import xml2db as _xml2db, db2xml as _db2xml


# =============================================================================
def includeme(configurator):
    """Function to register the module.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    # Registration
    Module.register(configurator, ModuleCioSkeleton)
    if not isinstance(configurator, Configurator):
        return

    # Permissions
    configurator.include('chrysalio.modules.cioskeleton.security')

    # Routes
    configurator.include('chrysalio.modules.cioskeleton.routes')

    # Views
    for theme in configurator.registry['themes']:
        theme = '/theme/{0}'.format(theme.lower()) if theme else ''
        path = '{0}/cioskeleton/images/'.format(theme)
        if configurator.introspector.get('static views', path) is None:
            configurator.add_static_view(
                path, join(dirname(__file__), 'Static', 'Images'))
        path = '{0}/cioskeleton/css/'.format(theme)
        if configurator.introspector.get('static views', path) is None:
            configurator.add_static_view(
                path, join(dirname(__file__), 'Static', 'Css'))
    configurator.scan('chrysalio.modules.cioskeleton.views')


# =============================================================================
class ModuleCioSkeleton(Module):
    """Class for an example of Chrysalio module.

    :param str config_ini:
        Absolute path to the configuration file (e.g. development.ini).
    """

    name = _('Skeleton')
    implements = ('skeleton',)
    dependencies = ('chrysalio.includes.themes',)
    relaxng = RELAXNG_CIOSKELETON
    xml2db = (_xml2db,)
    db2xml = (_db2xml,)
    _DBModule = DBModule

    # -------------------------------------------------------------------------
    def activate(self, registry, dbsession):
        """Method to activate the module.

        :type  registry: pyramid.registry.Registry
        :param registry:
            Application registry.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        """
        if registry['principals'][0][0] == 'mode' and \
           PRINCIPALS_CIOSKELETON[0][2][0] not in registry['principals'][0][2]:
            registry['principals'][0][2].append(
                PRINCIPALS_CIOSKELETON[0][2][0])
        if PRINCIPALS_CIOSKELETON[1] not in registry['principals']:
            registry['principals'].append(PRINCIPALS_CIOSKELETON[1])

        if 'modes' in registry and MODE_CIOSKELETON not in registry['modes']:
            registry['modes'].insert(1, MODE_CIOSKELETON)
        elif 'menu' in registry and MENU_CIOSKELETON not in registry['menu']:
            registry['menu'].insert(1, MENU_CIOSKELETON)

    # -------------------------------------------------------------------------
    def deactivate(self, registry, dbsession):
        """Method to deactivate the module.

        :type  registry: pyramid.registry.Registry
        :param registry:
            Application registry.
        :type  dbsession: sqlalchemy.orm.session.Session
        :param dbsession:
            SQLAlchemy session.
        """
        if registry['principals'][0][0] == 'mode' and \
           PRINCIPALS_CIOSKELETON[0][2][0] in registry['principals'][0][2]:
            registry['principals'][0][2].remove(
                PRINCIPALS_CIOSKELETON[0][2][0])
        if PRINCIPALS_CIOSKELETON[1] in registry['principals']:
            registry['principals'].remove(PRINCIPALS_CIOSKELETON[1])

        if 'modes' in registry and MODE_CIOSKELETON in registry['modes']:
            registry['modes'].remove(MODE_CIOSKELETON)
        elif 'menu' in registry and MENU_CIOSKELETON in registry['menu']:
            registry['menu'].remove(MENU_CIOSKELETON)



"""Function to import and export database from and into XML files."""

from os.path import join, basename, exists, splitext, relpath
from tempfile import mkdtemp
from shutil import rmtree
from zipfile import ZIP_DEFLATED, ZipFile, LargeZipFile, BadZipfile
from cgi import FieldStorage

from lxml import etree
from sqlalchemy.exc import ProgrammingError

from pyramid.response import Response, FileResponse

from ..lib.i18n import _
from ..lib.utils import scandir, walk
from ..lib.xml import load_xml, validate_xml, create_entire_xml
from ..lib.xml import relaxng4validation
from .dbsettings import DBSettings
from .dbprofile import DBProfile
from .dbuser import DBUser
from .dbgroup import DBGroup


# =============================================================================
def xml2db(dbsession, tree, only=None, error_if_exists=True, modules=None):
    """Load an XML configuration file into the database.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  tree: lxml.etree.ElementTree
    :param tree:
        Content of the XML document.
    :param str only: (optional)
        If not ``None``, only the items of type ``only`` are loaded.
    :param bool error_if_exists: (default=True)
        It returns an error if an item already exists.
    :type  modules: collections.OrderedDict
    :param modules: (optional)
        Dictionary of modules to use to complete the loading.
    :rtype: list
    :return:
        A list of error messages.
    """
    # Load settings
    if only is None or only == 'settings':
        DBSettings.xml2db(dbsession, tree.find('settings'))

    # Load profiles
    errors = element2db(dbsession, tree, only, error_if_exists, {
        'tag': 'profile', 'class': DBProfile})
    profiles = [k[0] for k in dbsession.query(DBProfile.profile_id)]

    # Load users
    errors += element2db(dbsession, tree, only, error_if_exists, {
        'tag': 'user', 'class': DBUser, 'kwargs': {'profiles': profiles}})
    users = dict(dbsession.query(DBUser.login, DBUser.user_id))

    # Load groups
    errors += element2db(dbsession, tree, only, error_if_exists, {
        'tag': 'group', 'class': DBGroup,
        'kwargs': {'profiles': profiles, 'users': users}})

    # Load tables from modules
    errors += module_xml2db(dbsession, tree, only, error_if_exists, modules)

    return [k for k in errors if k is not None]


# =============================================================================
def element2db(dbsession, tree, only, error_if_exists, element):
    """Load XML elements into the database according to an XPath.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  tree: lxml.etree.ElementTree
    :param tree:
        Content of the configuration file.
    :param str only:
        If not ``None``, only the items of type ``only`` are loaded.
    :param bool error_if_exists:
        It returns an error if an item already exists.
    :param dict element:
        A dictionary with keys ``tag``, ``class`` and, possibly, ``kwargs``,
        ``relaxng`` representing the element to insert into the database.
    :rtype: list
    :return:
        A list of error messages.
    """
    if only is not None and only != element['tag']:
        return []

    # Without namespace
    errors = []
    relaxng = element.get('relaxng')
    namespace = relaxng.get('namespace') if relaxng is not None else None
    if namespace is None:
        for elt in tree.xpath('{0}|{0}s/{0}'.format(element['tag'])):
            errors.append(element['class'].xml2db(
                dbsession, elt, error_if_exists, element.get('kwargs')))
        return errors

    # With namespace
    must_validate = tree.xpath('namespace-uri()') != namespace
    for root_elt in tree.xpath(
            '//ns0:{0}'.format(relaxng['root']),
            namespaces={'ns0': namespace}):
        if must_validate:
            error = validate_xml(
                etree.ElementTree(root_elt), relaxng4validation(relaxng))
            if error is not None:
                errors.append(error)
                continue
        for elt in root_elt.xpath(
                './/ns0:{0}s/ns0:{0}|ns0:{0}'.format(element['tag']),
                namespaces={'ns0': namespace}):
            errors.append(element['class'].xml2db(
                dbsession, elt, error_if_exists, element.get('kwargs')))

    return [k for k in errors if k is not None]


# =============================================================================
def module_xml2db(dbsession, tree, only, error_if_exists, modules):
    """Load elements from included modules.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  tree: lxml.etree.ElementTree
    :param tree:
        Content of the configuration file.
    :param str only:
        If not ``None``, only the items of type ``only`` are loaded.
    :param bool error_if_exists:
        It returns an error if an item already exists.
    :type  modules: :class:`collections.OrderedDict` or ``None``
    :param modules:
        Dictionary of modules to use to complete the loading.
    :rtype: list
    :return:
        A list of error messages.
    """
    if modules is None:
        return []

    errors = []
    for module_id in modules:
        errors += modules[module_id].module_xml2db(
            dbsession, tree, only, error_if_exists)
    return errors


# =============================================================================
def db2xml(dbsession, modules=None):
    """Return a list of XML elements.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  modules: collections.OrderedDict
    :param modules: (optional)
        Dictionary of modules to use to complete the loading.
    :rtype: list
    """
    elements = []

    # General settings
    elements.append(DBSettings.db2xml(dbsession))

    # Profiles
    for dbprofile in dbsession.query(DBProfile).order_by('profile_id'):
        elements.append(dbprofile.db2xml(dbsession))

    # Users
    for dbuser in dbsession.query(DBUser).order_by('login'):
        if dbuser.user_id != 1 or dbuser.status != 'administrator':
            elements.append(dbuser.db2xml(dbsession))

    # Groups
    for dbgroup in dbsession.query(DBGroup).order_by('group_id'):
        elements.append(dbgroup.db2xml(dbsession))

    # Save tables of included modules
    elements += module_db2xml(dbsession, modules)

    return [k for k in elements if k is not None]


# =============================================================================
def module_db2xml(dbsession, modules):
    """Return a list of XML elements from included modules.

    :type  dbsession: sqlalchemy.orm.session.Session
    :param dbsession:
        SQLAlchemy session.
    :type  modules: :class:`collections.OrderedDict` or ``None``
    :param modules:
        Dictionary of modules to use to complete the loading.
    :rtype: list
    """
    if modules is None:
        return []

    elements = []
    for module_id in modules:
        elements += modules[module_id].module_db2xml(dbsession)
    return elements


# =============================================================================
def web2db(request, _xml2db, only=None, relaxngs=None, error_if_exists=True):
    """Read XML or ZIP files from Web and load them into the database.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param _xml2db:
        Function to load an XML configuration file.
    :param str only: (optional)
        If not ``None``, only the items of type ``only`` are loaded.
    :param dict relaxngs: (optional)
        Dictionary of extra Relax NG files.
    :param bool error_if_exists: (default=True)
        It returns an error if an item already exists.
    """
    def _web2db_xml(filename, data):
        """Sub function to load an XML file."""
        tree = load_xml(filename, relaxngs, data=data)
        # pylint: disable = protected-access
        if not isinstance(tree, etree._ElementTree):
            errors.append(tree)
            return
        # pylint: enable = protected-access
        try:
            errors.extend(_xml2db(
                request.dbsession, tree, only=only,
                error_if_exists=error_if_exists,
                modules=request.registry.get('modules')))
        except ProgrammingError as error:  # pragma: nocover
            errors.append(error)

    # RelaxNG
    relaxngs = {} if relaxngs is None else relaxngs
    relaxngs.update(relaxng4validation(request.registry['relaxng']))

    # Browse files
    errors = []
    attachments = request.registry.settings.get('attachments')
    for field_storage in request.POST.getall('file'):
        if not isinstance(field_storage, FieldStorage):
            continue

        # XML file
        ext = splitext(field_storage.filename)[1]
        if ext == '.xml':
            _web2db_xml(field_storage.filename, field_storage.file.read())
            continue

        # Not a ZIP file
        if ext != '.zip':
            errors.append(_('Format must be ZIP or XML!'))
            continue

        # ZIP file
        try:
            with ZipFile(field_storage.file, 'r') as zip_file:
                for zip_info in zip_file.infolist():
                    if splitext(zip_info.filename)[1] == '.xml':
                        _web2db_xml(
                            zip_info.filename,
                            zip_file.read(zip_info.filename))
                    elif attachments:
                        zip_file.extract(zip_info, attachments)
        except BadZipfile:
            errors.append(_('It is not a ZIP file!'))
            continue

    # Push errors into session
    for err in errors:
        request.session.flash(err, 'alert')


# =============================================================================
def db2web(request, dbitems, filename, relaxng=None):
    """Convert SqlAlchemy items into an XML file and its attachments embedded
    in a Pyramid response.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param list dbitems:
        A list of SqlAlchemy objects with a method db2xml() and a
        method attachments2directory().
    :param str filename:
        The name of the downloaded file.
    :param dict relaxng: (optional)
        A dictionary describing the Relax NG used by
        func:`.lib.xml.create_entire_xml`.
    :rtype: pyramid.response.Response
    :return:
        An object :class:`~pyramid.response.Response` or ``None``.
    """
    # pylint: disable = too-many-locals
    # Create the XML file
    root_elt = create_entire_xml(
        relaxng or request.registry['relaxng'],
        [k.db2xml(request.dbsession) for k in dbitems])
    # pylint: disable = protected-access
    if not isinstance(root_elt, etree._Element):  # pragma: nocover
        return Response(body=root_elt)
    # pylint: enable = protected-access

    # Collect attachments
    attachments = request.registry.settings.get('attachments')
    tmp_dir = None
    if attachments and exists(attachments):
        tmp_dir = mkdtemp(
            prefix=request.registry.settings['site.uid'],
            dir=request.registry.settings.get('temporary'))
        container = join(tmp_dir, 'Container')
        for dbitem in dbitems:
            dbitem.attachments2directory(attachments, container)

    # Return a XML file
    if not tmp_dir or not tuple(scandir(tmp_dir)):
        if tmp_dir:
            rmtree(tmp_dir)
        response = Response(
            body=etree.tostring(
                root_elt, pretty_print=True, encoding='utf-8',
                xml_declaration=True),
            content_type='application/xml')
        response.headerlist.append((
            'Content-Disposition', 'attachment; filename="{0}"'.format(
                filename)))
        return response

    # Return a ZIP file
    # pylint: disable = consider-using-with
    zip_file = ZipFile(
        join(tmp_dir, '{0}.zip'.format(splitext(filename)[0])), 'w',
        ZIP_DEFLATED)
    zip_file.writestr(
        filename, etree.tostring(
            root_elt, encoding='utf-8', xml_declaration=True,
            pretty_print=True))
    for root, unused_, files in walk(container):
        for name in files:
            name = join(root, name)
            try:
                zip_file.write(name, relpath(name, container))
            except LargeZipFile:  # pragma: nocover
                zip_file.close()
                rmtree(tmp_dir)
                return Response(request.localizer.translate(
                    _('This file is too big!')))
            except IOError as error:  # pragma: nocover
                zip_file.close()
                rmtree(tmp_dir)
                return Response(error)
    filename = zip_file.filename
    zip_file.close()

    response = FileResponse(
        filename, request=request, content_type='application/zip')
    response.headers['Content-Disposition'] = \
        'attachment; filename="{0}"'.format(basename(filename))
    rmtree(tmp_dir)
    return response



"""Panel management view callables."""

from pyramid.view import view_config


# =============================================================================
@view_config(route_name='panel_open', renderer='json', xhr=True)
@view_config(route_name='panel_close', renderer='json', xhr=True)
def panel_view(request):
    """Toggle the state of a panel.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """
    if 'panels' not in request.registry:
        return {}
    panel_id = request.matchdict['panel_id']
    if panel_id not in request.registry['panels']:
        return {}

    if request.matched_route.name == 'panel_open':
        request.registry['panels'][panel_id].open(request)
    else:
        request.registry['panels'][panel_id].close(request)
    return {}



"""Mode management view callables."""

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound, HTTPNotFound

from ..lib.i18n import _
from ..lib.menu import Menu
from ..lib.modes import Modes


# =============================================================================
@view_config(route_name='mode')
def mode_view(request, uid='modes', menu_uid='menu'):
    """Mode redirection.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str uid: (default='modes')
        Unique ID of the modes management.
    :param str menu_uid: (default='menu')
        Unique ID of the menu management.
    """
    if uid not in request.registry or uid not in request.session:
        raise HTTPNotFound(comment=_('Modes are not activated.'))
    Menu.invalidate(request, menu_uid)
    modes = Modes(request, uid, request.registry.get(uid, ()))
    return HTTPFound(modes.select(request.matchdict['mode_id']))



"""The :func:`.main` function is called when the ``pserve`` command is invoked
against this application.
"""
from pyramid.config import Configurator

from .initialize import Initialize
from .lib.i18n import locale_negotiator
from .includes import load_includes


# =============================================================================
def main(global_config, **settings):
    """This function returns a Pyramid WSGI application.

    :param dict global_config:
        Dictionary describing the INI file with keys ``__file__`` and ``here``.
    :param dict settings:
        Application settings of [app:main] section of INI file.
    :rtype: pyramid.router.Router
    :return:
        WSGI application.
    """
    configurator = Configurator(
        settings=settings,
        locale_negotiator=locale_negotiator,
        default_permission='authenticated')

    # Chrysalio configuration
    configurator.include('pyramid_chameleon')
    configurator.include('.relaxng')
    configurator.include('.models')
    configurator.include('.routes')
    configurator.include('.security')
    configurator.include('.subscribers')
    configurator.scan('.views')
    Initialize(configurator).complete(
        global_config, __package__, 'ciopopulate')
    configurator.commit()

    # Includes loading
    load_includes(configurator)

    return configurator.make_wsgi_app()



"""Home view callables."""

from pyramid.httpexceptions import HTTPFound

from ..lib.i18n import _
from ..lib.form import button
from ..lib.panel import Panel
from ..lib.menu import Menu
from ..lib.modes import Modes


# =============================================================================
def home_view(request, title=None, modes_uid='modes', menu_uid='menu'):
    """Home view.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str title: (optional)
        Alternative title of the home page.
    :param str modes_uid: (default='modes')
        Unique ID of the modes management.
    :param str menu_uid: (default='menu')
        Unique ID of the menu management.
    """
    # Mode
    mode_id = request.params.get('mode')
    if mode_id and modes_uid in request.registry \
       and modes_uid in request.session:
        Menu.invalidate(request, menu_uid)
        Modes(request, modes_uid, request.registry[modes_uid]).select(mode_id)

    # Route
    route_id = request.params.get('route')
    if route_id:
        return HTTPFound(request.route_path(route_id, **request.params))

    # Panel
    Panel.manage_panels(request)

    request.breadcrumbs(title or _('Home'), 1)
    return {'button': button}



# pylint: disable = import-outside-toplevel
"""Tests of ``modules`` classes and functions."""

from collections import OrderedDict

from lxml import etree

from pyramid.i18n import TranslationString
from pyramid.httpexceptions import HTTPForbidden

from . import DBUnitTestCase
from ..includes.modules.models import DBModule


# =============================================================================
class UModulesModule(DBUnitTestCase):
    """Unit test class for :class:`modules.Module`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        from ..security import PRINCIPALS

        super(UModulesModule, self).setUp()
        self.configurator.registry['modules'] = OrderedDict()
        self.configurator.registry['modules_off'] = set()
        self.configurator.registry['principals'] = list(PRINCIPALS)

    # -------------------------------------------------------------------------
    def test_register(self):
        """[u:modules.Module.register]"""
        from . import TEST_INI
        from ..scripts import ScriptRegistry
        from ..security import PRINCIPALS
        from ..modules.cioskeleton import ModuleCioSkeleton

        self.configurator.get_settings()['__file__'] = TEST_INI
        ModuleCioSkeleton.register(self.configurator, ModuleCioSkeleton)
        self.assertEqual(len(self.configurator.registry['modules']), 1)

        registry = ScriptRegistry({'__file__': TEST_INI})
        registry['principals'] = list(PRINCIPALS)
        registry['modules'] = OrderedDict()
        ModuleCioSkeleton.register(registry, ModuleCioSkeleton)
        self.assertEqual(len(registry['modules']), 1)

    # -------------------------------------------------------------------------
    def test_check_conflicts(self):
        """[u:modules.Module.check_conflicts]"""
        from . import TEST_INI
        from ..modules import Module

        module1 = Module(TEST_INI)
        module1.implements = ('foo', 'bar')
        module2 = Module(TEST_INI)
        includes = ('chrysalio.includes.themes', 'module1', 'module2')
        modules = OrderedDict((('module1', module1), ('module2', module2)))

        implementations, error = Module.check_conflicts(includes, modules)
        self.assertIsNone(error)
        self.assertIn('foo', implementations)
        self.assertIn('bar', implementations)
        self.assertIn('module1', implementations)
        self.assertIn('module2', implementations)

        module2.implements = ('bar',)
        implementations, error = Module.check_conflicts(includes, modules)
        self.assertIsInstance(error, TranslationString)
        self.assertEqual(error.mapping['m1'], 'module1')
        self.assertEqual(error.mapping['m2'], 'module2')
        self.assertEqual(error.mapping['f'], 'bar')
        self.assertFalse(implementations)

    # -------------------------------------------------------------------------
    def test_check_dependencies(self):
        """[u:modules.Module.check_dependencies]"""
        from . import TEST_INI
        from ..modules import Module

        module = Module(TEST_INI)

        self.assertIsNone(module.check_dependencies(
            ('chrysalio.includes.themes',)))

        module.dependencies = ('chrysalio.includes.themes',)
        self.assertIsNone(module.check_dependencies(
            ('chrysalio.includes.themes', 'chrysalio.modules.foo')))
        self.assertIsNone(module.check_dependencies(
            ('chrysalio.includes.themes',)))
        self.assertIsNotNone(module.check_dependencies(
            ('chrysalio.modules.foo',)))

        module.dependencies = (
            'chrysalio.includes.themes', 'bar-implementation')
        self.assertIsNone(module.check_dependencies(
            ('chrysalio.includes.themes', 'bar-implementation')))
        self.assertIsNotNone(module.check_dependencies(
            ('chrysalio.includes.themes',)))

    # -------------------------------------------------------------------------
    def test_check_activations(self):
        """[u:modules.Module.check_activations]"""
        from . import TEST_INI
        from ..modules import Module

        modules_off = set(['chrysalio.modules'])
        module = Module(TEST_INI)
        self.assertFalse(module.check_activations(modules_off))

        modules_off = set(['chrysalio.modules.cioskeleton'])
        module.dependencies = ('chrysalio.modules.cioskeleton',)
        self.assertTrue(module.check_activations(modules_off))

    # -------------------------------------------------------------------------
    def test_module_xml2db(self):
        """[u:modules.Module.module_xml2db]"""
        from . import TEST_INI
        from ..modules import Module

        def _xml2db(dbsession, root_elt, only, error_if_exists):
            """Dummy xml2db function."""
            # pylint: disable = unused-argument
            return []

        dbsession = self.request.dbsession
        tree = etree.XML(
            '<chrysalio version="1.0">'
            '  <module id="chrysalio.modules.cioskeleton" inactive="true"/>'
            '</chrysalio>')
        module = Module(TEST_INI)
        module.uid = 'chrysalio.modules.cioskeleton'

        # Without _DBModule and settings
        self.assertFalse(module.module_xml2db(dbsession, tree, None, False))
        self.assertIsNone(dbsession.query(DBModule).filter_by(
            module_id=module.uid).first())

        # Inactive module
        # pylint: disable = protected-access
        module._DBModule = DBModule
        # pylint: enable = protected-access
        errors = module.module_xml2db(dbsession, tree, None, False)
        self.assertFalse(errors)
        self.assertIsNotNone(dbsession.query(DBModule).filter_by(
            module_id=module.uid).first())

        # Active module
        tree = etree.XML(
            '<chrysalio version="1.0">'
            '  <module id="chrysalio.modules.cioskeleton" inactive="false"/>'
            '</chrysalio>')
        self.assertFalse(module.module_xml2db(dbsession, tree, None, False))
        self.assertIsNone(dbsession.query(DBModule).filter_by(
            module_id=module.uid).first())

        # Without settings
        tree = etree.XML(
            '<chrysalio version="1.0">'
            '  <module id="foo" inactive="true"/>'
            '</chrysalio>')
        module.relaxng = {
            'namespace': 'http://ns.chrysal.io/chrysalio/skeleton',
            'root': 'chrysalio-skeleton'}
        module.xml2db = (_xml2db,)
        self.assertFalse(module.module_xml2db(dbsession, tree, None, False))

        # With settings in a backup
        tree = etree.XML(
            '<chrysalio version="1.0">'
            '  <module id="chrysalio.modules.cioskeleton" inactive="true">'
            '    <chrysalio-skeleton'
            '        xmlns="http://ns.chrysal.io/chrysalio/skeleton"'
            '        version="1.0">'
            '      <bones>'
            '        <bone id="hand"><label>Hand</label></bone>'
            '      </bones>'
            '    </chrysalio-skeleton>'
            '  </module>'
            '</chrysalio>')
        self.assertFalse(module.module_xml2db(dbsession, tree, None, False))

        # With settings in a simple configuration
        tree = etree.XML(
            '<chrysalio-skeleton'
            '    xmlns="http://ns.chrysal.io/chrysalio/skeleton"'
            '    version="1.0">'
            '  <bones>'
            '    <bone id="hand"><label>Hand</label></bone>'
            '  </bones>'
            '</chrysalio-skeleton>')
        self.assertFalse(module.module_xml2db(dbsession, tree, None, False))

    # -------------------------------------------------------------------------
    def test_module_db2xml(self):
        """[u:modules.Module.module_db2xml]"""
        from . import TEST_INI
        from ..modules import Module

        def _db2xml(dbsession, root_elt):
            """Dummy db2xml function."""
            # pylint: disable = unused-argument, unnecessary-pass
            pass

        dbsession = self.request.dbsession
        module = Module(TEST_INI)
        module.uid = 'chrysalio.modules.cioskeleton'

        # Without _DBModule
        elements = module.module_db2xml(dbsession)
        self.assertEqual(len(elements), 0)

        # Inactive, without settings
        dbsession.add(DBModule(module_id=module.uid, inactive=True))
        # pylint: disable = protected-access
        module._DBModule = DBModule
        # pylint: enable = protected-access
        elements = module.module_db2xml(dbsession)
        self.assertEqual(len(elements), 1)
        self.assertEqual(elements[0].tag, 'module')
        self.assertEqual(elements[0].get('inactive'), 'true')

        # Without settings
        module.relaxng = {
            'namespace': 'http://ns.chrysal.io/chrysalio/skeleton',
            'root': 'chrysalio-skeleton', 'version': '1.0'}
        module.db2xml = (_db2xml,)
        elements = module.module_db2xml(dbsession)
        self.assertEqual(len(elements), 1)
        self.assertEqual(len(elements[0]), 1)
        self.assertEqual(
            elements[0][0].tag,
            '{http://ns.chrysal.io/chrysalio/skeleton}chrysalio-skeleton')

    # -------------------------------------------------------------------------
    def test_poppulate(self):
        """[u:modules.Module.populate]"""
        from . import TEST_INI
        from ..modules import Module

        dbsession = self.request.dbsession
        self.assertIsNone(Module(TEST_INI).populate({}, {}, dbsession))

    # -------------------------------------------------------------------------
    def test_backup(self):
        """[u:modules.Module.backup]"""
        from . import TEST_INI, BACKUP_DIR
        from ..modules import Module

        dbsession = self.request.dbsession
        self.assertIsNone(
            Module(TEST_INI).backup({}, {}, dbsession, BACKUP_DIR))

    # -------------------------------------------------------------------------
    def test_activate(self):
        """[u:modules.Module.activate]"""
        from . import TEST_INI
        from ..modules import Module

        dbsession = self.request.dbsession
        self.assertIsNone(Module(TEST_INI).activate({}, dbsession))

    # -------------------------------------------------------------------------
    def test_deactivate(self):
        """[u:modules.Module.deactivate]"""
        from . import TEST_INI
        from ..modules import Module

        dbsession = self.request.dbsession
        self.assertIsNone(Module(TEST_INI).deactivate({}, dbsession))

    # -------------------------------------------------------------------------
    def test_check_activated(self):
        """[u:modules.Module.check_activated]"""
        from ..modules import Module

        self.assertRaises(
            HTTPForbidden, Module.check_activated, self.request,
            'chrysalio.modules.cioskeleton')

    # -------------------------------------------------------------------------
    def test_configuration_route(self):
        """[u:modules.Module.configuration_route]"""
        from . import TEST_INI
        from ..modules import Module

        self.assertIsNone(Module(TEST_INI).configuration_route(self.request))

    # -------------------------------------------------------------------------
    def test_settings(self):
        """[u:modules.Module._settings]"""
        from . import TEST_INI
        from ..modules import Module

        # With standard section
        # pylint: disable = protected-access
        settings = Module(TEST_INI)._settings(TEST_INI)
        # pylint: enable = protected-access
        self.assertIsInstance(settings, dict)
        self.assertIn('here', settings)

        # With CioFoo section
        # pylint: disable = protected-access
        settings = Module(TEST_INI)._settings(TEST_INI, 'CioFoo')
        # pylint: enable = protected-access
        self.assertIn('here', settings)
        self.assertIn('bar', settings)



# pylint: disable = import-outside-toplevel
"""Tests of ``lib.attachment`` functions."""

from os.path import exists, join, dirname
from unittest import TestCase
from cgi import FieldStorage

from pyramid import testing
from pyramid.testing import DummyRequest

from . import TmpDirTestCase


# =============================================================================
class ULibAttachmentAttachmentUrl(TestCase):
    """Unit test class for :func:`lib.attachment.attachment_url`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.attachment.attachment_url]"""
        from ..lib.attachment import attachment_url

        request = testing.DummyRequest()
        request.registry.settings = {}
        self.assertIsNone(
            attachment_url(request, 'Users', None, 'test5.svg'))
        self.assertIsNone(
            attachment_url(request, 'Users', 'Test5', None))
        self.assertIsNone(
            attachment_url(request, 'Users', 'Test5', 'test5.svg'))

        configurator = testing.setUp(settings={
            'attachments': join(dirname(__file__), 'Attachments')})
        configurator.add_route('attachment', '/attachment/*path')
        self.assertIsNone(
            attachment_url(request, 'Users', 'Test5', 'test4.svg'))
        self.assertEqual(
            attachment_url(request, 'Users', 'Test5', 'test5.svg'),
            '/attachment/Users/Test5/test5.svg')


# =============================================================================
class ULibAttachmentAttachmentUpdate(TmpDirTestCase):
    """Unit test class for :func:`lib.attachment.attachment_update`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.attachment.attachment_update]"""
        from . import TEST_DIR, TEST1_SVG
        from ..lib.attachment import attachment_update

        request = DummyRequest()
        request.registry.settings = {}
        attachments_key, file_path = attachment_update(
            request, 'Users', None, FieldStorage(), 'test1.svg')
        self.assertIsNone(attachments_key)
        self.assertEqual(file_path, 'test1.svg')

        request.registry.settings = {'attachments': TEST_DIR}
        with open(TEST1_SVG, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST1_SVG
            input_file.file = hdl
            attachments_key, file_path = attachment_update(
                request, 'Users', None, input_file)
        self.assertFalse(request.session.pop_flash('alert'))
        self.assertIsNotNone(attachments_key)
        self.assertNotEqual(attachments_key, 'Test1')
        self.assertIsNotNone(file_path)
        self.assertTrue(exists(join(
            TEST_DIR, 'Users', attachments_key, file_path)))

        with open(TEST1_SVG, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST1_SVG
            input_file.file = hdl
            attachments_key2, file_path2 = attachment_update(
                request, 'Users', attachments_key, input_file, file_path)
        self.assertEqual(attachments_key, attachments_key2)
        self.assertNotEqual(file_path, file_path2)
        self.assertFalse(exists(join(
            TEST_DIR, 'Users', attachments_key2, file_path)))
        self.assertTrue(exists(join(
            TEST_DIR, 'Users', attachments_key2, file_path2)))

        with open(TEST1_SVG, 'rb') as hdl:
            input_file = FieldStorage()
            input_file.filename = TEST1_SVG
            input_file.file = hdl
            attachments_key, file_path = attachment_update(
                request, 'Users', 'Test1', input_file, file_path)
        self.assertEqual(attachments_key, 'Test1')
        self.assertIsNotNone(file_path)



"""Class to manage modes."""

from webhelpers2.html import literal
from pyramid.httpexceptions import HTTPNotFound, HTTPForbidden

from ..includes.themes import theme_static_prefix
from ..lib.i18n import _


# =============================================================================
class Modes(object):
    """Modes base class.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    :param str uid:
        Unique ID used to store the mode in the session.
    :param list framework:
        Full version of the mode.

    The framework of the mode is a list of available modes. Each item of the
    list has the following structure:

        mode = (mode_id, (icon, label, permission, route, menu))

    ``route`` can be a name of route, a tuple such as
    ``(route_name, params_dict)`` or a function returning such a tuple.

    ``menu`` can be a list of menu entries or a function returning such a
    list.

    In the user session, under ``uid`` key, modes are stored as a list
    ``[current_mode_id, current_label, modes]``. ``modes`` is a list of modes
    filtered according to user permissions.
    """

    # -------------------------------------------------------------------------
    def __init__(self, request, uid, framework):
        """Constructor method."""
        self._request = request
        self.uid = uid
        self._framework = framework

    # -------------------------------------------------------------------------
    def label(self):
        """Return the translated label of the current mode.

        :rtype: str
        """
        self._construct()
        return self._request.session[self.uid][1]

    # -------------------------------------------------------------------------
    def menu_framework(self):
        """Return the framework of the menu.

        :rtype: tuple
        """
        if self.uid not in self._request.session:
            mode = self._framework[0]
            return mode[1][4][4] or () if mode[1][4] else ()

        current_mode_id = self._request.session[self.uid][0]
        for mode in self._framework:
            if mode[0] == current_mode_id:
                return mode[1][4][4] or () if mode[1][4] else ()
        return ()

    # -------------------------------------------------------------------------
    def select(self, mode_id):
        """Select a new mode.

        :param str mode_id:
            ID of menu to select.
        :rtype: str
        :return:
            Return the default route for the selected mode.
        """
        mode = dict(self._framework).get(mode_id)
        if mode is None:
            raise HTTPNotFound(comment=_('Unknown mode ${m}.', {'m': mode_id}))

        if mode[2] is not None and not self._request.has_permission(mode[2]):
            raise HTTPForbidden()

        self._request.session[self.uid] = [
            mode_id, self._request.localizer.translate(mode[1]), None]
        self._construct()

        route = mode[3] if isinstance(mode[3], (list, tuple)) \
            else (mode[3], {})
        return self._request.route_path(route[0], **route[1])

    # -------------------------------------------------------------------------
    def xhtml(self):
        """Return a <ul> structure with the list of modes.

        :rtype: webhelpers2.html._literal.literal
        """
        self._construct()
        return self._request.session[self.uid][2]

    # -------------------------------------------------------------------------
    @classmethod
    def invalidate(cls, request, uid):
        """Invalidate the mode ``uid`` if exists.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str uid:
            Unique ID of the mode to invalidate.
        """
        if uid in request.session:
            request.session[uid][2] = None

    # -------------------------------------------------------------------------
    def _construct(self):
        """If the mode does not exist in session, construct it according to
        user permissions and put it in ``session[self.uid]``."""
        # Mode in session
        if self.uid in self._request.session and \
           self._request.session[self.uid][2] is not None:
            return

        # List of modes
        theme = theme_static_prefix(self._request)
        translate = self._request.localizer.translate
        current_mode_id = None
        current_label = None
        authorized = set()
        html = ''
        for mode in self._framework:
            if not self._has_permission(mode):
                continue
            label = translate(mode[1][1])
            if current_mode_id is None:
                current_mode_id = mode[0]
                current_label = label
            html += '<li><span><a href="{url}">'.format(
                url=self._request.route_path('mode', mode_id=mode[0]))
            if mode[1][0]:  # icon
                html += '<img src="{icon}" alt="{label}"'\
                    ' title="{label}"/> '.format(
                        icon=mode[1][0].format(theme=theme), label=label)
            html += '{label}</a></span></li>'.format(label=label)
            authorized.add(mode[0])
        if not html:
            self._request.session[self.uid] = [None, '', html]
            return
        html = '<ul id="cioModes-{uid}">{html}</ul>'.format(
            uid=self.uid, html=html) if len(authorized) > 1 else ''

        # Default mode
        if self.uid in self._request.session and \
           self._request.session[self.uid][0] in authorized:
            current_mode_id = self._request.session['modes'][0]
            current_label = self._request.session['modes'][1]

        # Session
        self._request.session[self.uid] = [
            current_mode_id, current_label, literal(html)]

    # -------------------------------------------------------------------------
    def _has_permission(self, mode):
        """Check if the user has the permission to use this mode.

        :param tuple mode:
            The mode to check.
        :rtype: bool
        """
        if mode[1][2] is None:
            return True
        return self._request.has_permission(mode[1][2]).boolval



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.paging`` function and classes."""

from re import match
from collections import namedtuple

from webhelpers2.html import literal

# pylint: disable = unused-import
from ..models.dbgroup import DBGroupUser  # noqa
# pylint: enable = unused-import
from . import DBUnitTestCase


# =============================================================================
class ULibPagingSortableColumn(DBUnitTestCase):
    """Unit test class for :func:`lib.paging.sortable_column`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.paging.sortable_column]"""
        from ..lib.paging import sortable_column

        self.configurator.add_route('user_index', '/usr/index')
        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_index')
        self.request.GET = {'quick': 1}

        html = sortable_column(
            self.request, 'Foo Column', 'login', current_sorting='+login')
        self.assertIsInstance(html, literal)
        self.assertIn('Foo Column', html)
        self.assertIn('quick=1', html)
        self.assertIn('sort=-login', html)
        self.assertNotIn('sort=+login', html)


# =============================================================================
class ULibPagingPaging(DBUnitTestCase):
    """Unit test class for :class:`lib.paging.Paging`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        super(ULibPagingPaging, self).setUp()
        self.configurator.add_route('user_index', '/usr/index')
        self.request.matched_route = namedtuple('Route', 'name')(
            name='user_index')

        self.add_user({
            'login': 'test1', 'last_name': 'AVULEUR', 'first_name': 'Édith',
            'password': 'test1pwd', 'email': 'test1@chrysal.io'})
        self.add_user({
            'login': 'test2', 'first_name': 'Sophie', 'last_name': 'FONFEC',
            'status': 'inactive', 'password': 'test2pwd',
            'email': 'test2@chrysal.io'})
        self.add_user({
            'login': 'test3', 'first_name': 'Guy', 'last_name': 'LIGUILI',
            'status': 'locked',
            'password': 'test3pwd', 'email': 'test3@chrysal.io'})

    # -------------------------------------------------------------------------
    def test_init(self):
        """[u:lib.paging.Paging.__init__]"""
        from ..models.dbuser import DBUser
        from ..lib.paging import PAGE_DEFAULT_SIZE, Paging

        self.assertRaises(TypeError, Paging, None, 'users', 12, item_count=1)

        paging = Paging(None, 'users1', ('user1', 'user2'))
        self.assertEqual(paging.paging_id, 'users1')
        self.assertEqual(paging.page_size, PAGE_DEFAULT_SIZE)
        self.assertEqual(paging.item_count, 2)
        self.assertEqual(paging.page_count, 1)

        # pylint: disable = no-member
        query = self.request.dbsession.query(DBUser.user_id, DBUser.login)
        # pylint: enable = no-member
        paging = Paging(None, 'users2', query)
        self.assertEqual(paging.paging_id, 'users2')
        self.assertEqual(paging.item_count, 3)
        self.assertEqual(paging.page_count, 1)

        self.request.params = {
            'page_size': '2', 'sort': 'login', 'display': 'cards', 'page': '3'}
        paging = Paging(self.request, 'users3', query)
        self.assertEqual(paging.page_size, 2)
        self.assertEqual(paging.page, 2)

        self.request.params = {'page_size': '0'}
        paging = Paging(self.request, 'users4', query)
        self.assertEqual(paging.page_size, 0)
        self.assertEqual(paging.page, 1)
        self.assertEqual(paging.item_count, 3)
        self.assertEqual(paging.page_count, 1)

    # -------------------------------------------------------------------------
    def test_params(self):
        """[u:lib.paging.Paging.params]"""
        from ..lib.paging import PAGE_DEFAULT_SIZE, Paging

        params = Paging.params(None, 'users', '+login')
        self.assertEqual(len(params), 4)
        self.assertIn('page_size', params)
        self.assertEqual(params['page_size'], PAGE_DEFAULT_SIZE)

        self.request.registry['settings']['page-size'] = 40
        params = Paging.params(self.request, 'users', '+login')
        self.assertIn('page_size', params)
        self.assertEqual(params['page_size'], 40)

        self.request.POST = {'page_size': '160'}
        self.request.params = self.request.POST
        params = Paging.params(self.request, 'users', '+login')
        self.assertIn('sort', params)
        self.assertEqual(params['sort'], '+login')
        self.assertIn('page_size', params)
        self.assertEqual(params['page_size'], 160)
        self.assertIn('display', params)
        self.assertEqual(params['display'], 'cards')
        self.assertIn('paging', self.request.session)
        self.assertEqual(self.request.session['paging'][0], 40)
        self.assertIn('users', self.request.session['paging'][1])
        paging = self.request.session['paging'][1]['users']
        self.assertEqual(paging['page'], 1)
        self.assertEqual(paging['sort'], '+login')

        self.request.POST = {'sort': '-login'}
        self.request.params = self.request.POST
        params = Paging.params(self.request, 'users', '+login')
        self.assertEqual(params['sort'], '-login')
        paging = self.request.session['paging'][1]['users']
        self.assertEqual(paging['sort'], '-login')

    # -------------------------------------------------------------------------
    def test_get_sort(self):
        """[u:lib.paging.Paging.get_sort]"""
        from ..lib.paging import Paging

        self.assertIsNone(Paging.get_sort(self.request, 'users'))

        Paging.params(self.request, 'users', '+login')
        self.assertEqual(Paging.get_sort(self.request, 'users'), '+login')

    # -------------------------------------------------------------------------
    def test_get_page(self):
        """[u:lib.paging.Paging.get_page]"""
        from ..lib.paging import Paging

        self.assertEqual(Paging.get_page(self.request, 'users'), 1)

        params = Paging.params(self.request, 'numbers')
        params['page'] = 2
        paging = Paging(self.request, 'numbers', range(100), params)
        self.assertEqual(paging.get_page(self.request, 'numbers'), 2)

    # -------------------------------------------------------------------------
    def test_get_item(self):
        """[u:lib.paging.Paging.get_item]"""
        from ..lib.paging import Paging

        paging = Paging(None, 'users1', ('user1', 'user2'))
        self.assertIsNone(paging.get_item('file_id', 'user1'))

        paging = Paging(None, 'files', (
            {'file_id': 'qsRz', 'name': 'foo'},
            {'file_id': 'sqdh', 'name': 'bar'},
            {'file_id': 'qsRz', 'name': 'baz'}))
        pitem = paging.get_item('file_id', 'qsRz')
        self.assertIsNotNone(pitem)
        self.assertIsInstance(pitem, dict)
        self.assertIn('name', pitem)
        self.assertEqual(pitem['name'], 'foo')

    # -------------------------------------------------------------------------
    def test_set_current_ids(self):
        """[u:lib.paging.Paging.set_current_ids]"""
        from ..models.dbuser import DBUser
        from ..lib.paging import Paging

        # pylint: disable = no-member
        query = self.request.dbsession.query(DBUser.user_id, DBUser.login)
        # pylint: enable = no-member

        # Without request
        paging = Paging(None, 'users', query)
        paging.set_current_ids('user_id')
        self.assertNotIn('paging', self.request.session)

        # With a SQLAlchemy query
        paging = Paging(self.request, 'users', query)
        paging.set_current_ids('user_id')
        self.assertIn('paging', self.request.session)
        self.assertIn('users', self.request.session['paging'][1])
        self.assertIn(
            'current_ids', self.request.session['paging'][1]['users'])
        self.assertEqual(
            len(self.request.session['paging'][1]['users']['current_ids']), 3)

        # With an empty collection
        paging = Paging(self.request, 'files', [])
        paging.set_current_ids('file_id')
        self.assertEqual(
            len(self.request.session['paging'][1]['files']['current_ids']), 0)

    # -------------------------------------------------------------------------
    def test_pager(self):
        """[u:lib.paging.Paging.pager]"""
        from ..models.dbuser import DBUser
        from ..lib.paging import Paging

        # pylint: disable = no-member
        dbquery = self.request.dbsession.query(DBUser.user_id, DBUser.login)
        # pylint: enable = no-member

        html = Paging(None, 'users', dbquery).pager()
        self.assertEqual(html, '')

        html = Paging(self.request, 'users', dbquery).pager()
        self.assertIn('<span>1</span>', html)

        html = Paging(
            self.request, 'users', dbquery,
            params={'page_size': 2, 'page': 1, 'display': 'cards'}).pager()
        self.assertIn('<span>1</span>', html)
        self.assertIn('page=2', html)

    # -------------------------------------------------------------------------
    def test_pager_top(self):
        """[u:lib.paging.Paging.pager_top]"""
        from ..models.dbuser import DBUser
        from ..lib.paging import Paging

        # pylint: disable = no-member
        dbquery = self.request.dbsession.query(
            DBUser.user_id, DBUser.login, DBUser.last_name)
        # pylint: enable = no-member

        html = Paging(None, 'users', dbquery).pager_top()
        self.assertEqual(html, '&nbsp;')

        html = Paging(self.request, 'users', dbquery).pager_top()
        self.assertIn('first_off', html)
        self.assertIn('previous_off', html)
        self.assertIn('next_off', html)
        self.assertIn('last_off', html)

        params = {'page_size': 2, 'page': 1, 'sort': None, 'display': 'cards'}
        html = Paging(self.request, 'users', dbquery, params).pager_top()
        self.assertIn('page=2', html)
        self.assertIn('src="/images/paging_go_next.png"', html)
        self.assertIn('src="/images/paging_go_last.png"', html)

    # -------------------------------------------------------------------------
    def test_pager_bottom(self):
        """[u:lib.paging.Paging.pager_bottom]"""
        from ..models.dbuser import DBUser
        from ..lib.paging import Paging

        # pylint: disable = no-member
        dbquery = self.request.dbsession.query(
            DBUser.user_id, DBUser.login, DBUser.last_name)
        # pylint: enable = no-member

        html = Paging(self.request, 'users', dbquery).pager_bottom()
        self.assertIn('<span>1</span>', html)

        params = {'page_size': 2, 'page': 1, 'sort': None, 'display': 'cards'}
        html = Paging(self.request, 'users', dbquery, params).pager_bottom()
        self.assertIn('page=2', html)

    # -------------------------------------------------------------------------
    def test_display_modes(self):
        """[u:lib.paging.Paging.display_mode]"""
        from ..lib.paging import Paging

        html = Paging(None, 'users', ['user1']).display_modes()
        self.assertEqual(html, '')

        self.request.GET = {'quick': 1}
        paging = Paging(self.request, 'users', ['user1'])
        html = paging.display_modes()
        self.assertNotIn('display=cards', html)
        self.assertIn('display=list', html)
        self.assertIn('quick=1', html)

        paging.display = 'list'
        html = paging.display_modes()
        self.assertIn('display=cards', html)
        self.assertNotIn('display=list', html)

    # -------------------------------------------------------------------------
    def test_sortable_column(self):
        """[u:lib.paging.Paging.sortable_column]"""
        from ..lib.paging import Paging

        html = Paging(None, 'users', ['user1']).sortable_column(
            'Foo Column', 'login')
        self.assertEqual(html, '&nbsp;')

        html = Paging(self.request, 'users', ['user1']).sortable_column(
            'Foo Column', 'login')
        self.assertIn('Foo Column', html)
        self.assertIn('sort=%2Blogin', html)
        self.assertNotIn('sort=-login', html)

    # -------------------------------------------------------------------------
    def test_navigator(self):
        """[u:lib.paging.Paging.navigator]"""
        from ..lib.paging import Paging

        # With an empty session
        html = Paging.navigator(
            self.request, 'profiles', 'main_manager', '/profile/view/_ID_')
        self.assertFalse(html)

        # With an item which is not in current IDs
        self.request.session['paging'] = (
            80, {'profiles': {'current_ids': [
                'main_manager', 'user_creator', 'user_editor']}})
        html = Paging.navigator(
            self.request, 'profiles', 'foo', '/profile/view/_ID_')
        self.assertFalse(html)

        # With the first item
        html = Paging.navigator(
            self.request, 'profiles', 'main_manager', '/profile/view/_ID_')
        self.assertIn('<span>1</span>', html)
        self.assertIn('user_creator', html)

        # With the last item
        html = Paging.navigator(
            self.request, 'profiles', 'user_editor', '/profile/view/_ID_')
        self.assertIn('<span>3</span>', html)
        self.assertIn('user_creator', html)

    # -------------------------------------------------------------------------
    def test_range(self):
        """[u:lib.paging.Paging._range]"""
        from ..lib.paging import Paging

        collection = [
            'user01', 'user02', 'user03', 'user04', 'user05', 'user06',
            'user07', 'user08', 'user09', 'user10', 'user11', 'user12']
        params = {'page_size': 2, 'page': 5, 'sort': None, 'display': 'cards'}
        paging = Paging(self.request, 'users', collection, params)

        # pylint: disable = protected-access
        html = paging._range(match(r'~(\d+)~', '~2~'))
        self.assertIn('page=1', html)
        self.assertIn('<span>5</span>', html)

        paging.page = 2
        html = paging._range(match(r'~(\d+)~', '~2~'))
        self.assertIn('<span>2</span>', html)



"""Cache functionality."""

from time import time

from beaker.cache import CacheManager, cache_regions
from beaker.util import parse_cache_config_options

from pyramid.config import Configurator


CACHE_DEFAULT_TTL = 43200
CREATED_KEY = '__created__'


# =============================================================================
def includeme(configurator):
    """Function to include cache functionnality based on registry for global
    cache and session for user cache.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    if isinstance(configurator, Configurator):
        if 'cache_user' not in configurator.registry:
            configurator.registry['cache_user'] = CacheUser(
                configurator.get_settings())
        if 'cache_global' not in configurator.registry:
            configurator.registry['cache_global'] = CacheGlobal(
                configurator.get_settings())


# =============================================================================
class CacheUser(object):
    """Class to manage a private cache for a user based on session."""

    # -------------------------------------------------------------------------
    def __init__(self, settings):
        """Constructor method."""
        prefix = ''
        regions = settings.get('cache.regions')
        if not regions:
            prefix = 'beaker.'
            regions = settings.get('beaker.cache.regions', '')
        self.regions = {}
        for region in regions.split(','):
            region = region.strip()
            expire = settings.get(
                '{0}cache.{1}.expire'.format(prefix, region))
            self.regions[region] = int(expire) if expire and expire.isdigit() \
                else CACHE_DEFAULT_TTL

    # -------------------------------------------------------------------------
    def set(self, request, key, value, namespace='', region=None, expire=None):
        """Set value into the cache.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str key:
            Key of value to retrieve.
        :param value:
            The value associated with the key or ``None``.
        :param str namespace: (default='')
            Cache namespace.
        :param str region: (optional)
            Beaker cache region.
        :param int expire: (optional)
            Special expiration.
        """
        # pylint: disable = too-many-arguments
        if 'cache' not in request.session:
            request.session['cache'] = {}
        cache = request.session['cache']
        if namespace not in cache:
            cache[namespace] = {}

        if not expire:
            expire = self.regions.get(region, CACHE_DEFAULT_TTL)
        cache[namespace][key] = (value, time() + expire)

    # -------------------------------------------------------------------------
    def get(self, request, key, namespace=''):
        """Get value from the cache.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str key:
            Key of value to retrieve.
        :param str namespace: (default='')
            Cache namespace.
        :return:
            The value associated with the key or ``None``.
        """
        if 'cache' not in request.session or \
           namespace not in request.session['cache'] or \
           key not in request.session['cache'][namespace]:
            return None

        if request.session['cache'][namespace][key][1] < time():
            del request.session['cache'][namespace][key]
            self.purge(request, namespace)
            return None

        return request.session['cache'][namespace][key][0]

    # -------------------------------------------------------------------------
    def clear(self, request, key=None, namespace=''):
        """Clear a key/value or the entire namespace ``namespace``.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str key: (optional)
            Key of value to remove.
        :param str namespace: (default='')
            Cache namespace.
        """
        if 'cache' not in request.session or \
           namespace not in request.session['cache']:
            return

        if key is not None:
            if key in request.session['cache'][namespace]:
                del request.session['cache'][namespace][key]
            self.purge(request, namespace)
            return

        del request.session['cache'][namespace]
        self.purge(request)

    # -------------------------------------------------------------------------
    @classmethod
    def purge(cls, request, namespace=None):
        """Purge expired entries.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str namespace: (optional)
            Cache namespace.
        """
        now = time()
        namespaces = (namespace,) if namespace is not None \
            else tuple(request.session['cache'])
        for nspace in namespaces:
            for key in tuple(request.session['cache'][nspace]):
                if request.session['cache'][nspace][key][1] < now:
                    del request.session['cache'][nspace][key]
            if not request.session['cache'][nspace]:
                del request.session['cache'][nspace]
        if not request.session['cache']:
            del request.session['cache']


# =============================================================================
def cache_user_access(namespace_prefix, region=None):
    """A decorator to retrieve in the user cache the result of an access
    method whose the two first arguments are ``request`` and ``item``.
    ``item`` must have a ``uid`` attribute.

    :param str namespace_prefix:
        Prefix of the cache namespace.
    :param str region: (optional)
        Name of region.
    """
    def _decorated(method):
        """Decoration of the method `method`."""

        def _wrapper(class_, request, item, *args, **kwargs):
            """Use of user cache."""
            access_method = method.__func__ \
                if isinstance(method, classmethod) else method
            if item is None:
                return access_method(class_, request, item, *args, **kwargs)
            if not hasattr(item, 'uid'):
                raise Exception('Class must have a "uid" attribute!')

            namespace = cache_namespace(namespace_prefix, item.uid)
            access = request.registry['cache_user'].get(
                request, 'access', namespace)
            if access is not None:
                return access

            access = access_method(class_, request, item, *args, **kwargs)
            request.registry['cache_user'].set(
                request, 'access', access, namespace, region)
            return access

        return _wrapper

    return _decorated


# =============================================================================
class CacheGlobal(object):
    """Class to manage a global cache based on Beaker cache.

    :param dict settings:
         Pyramid settings.
    """

    # -------------------------------------------------------------------------
    def __init__(self, settings):
        """Constructor method."""
        self._cache_manager = CacheManager(
            **parse_cache_config_options(settings))
        self.regions = cache_regions
        self._caches = {}

    # -------------------------------------------------------------------------
    def set(self, key, value, namespace='', region=None, **kwargs):
        """Set value into the cache.

        :param str key:
            Key of value to retrieve.
        :param value:
            The value associated with the key or ``None``.
        :param str namespace: (default='')
            Cache namespace.
        :param str region: (optional)
            Beaker cache region.
        :param dict kwargs:
            Keyworded arguments.
        """
        cache = self._get_cache(namespace, region, **kwargs)
        if cache is not None:
            cache.put(key, value)

    # -------------------------------------------------------------------------
    def get(self, key, namespace='', region=None, **kwargs):
        """Get value from the cache.

        :param str key:
            Key of value to retrieve.
        :param str namespace: (default='')
            Cache namespace.
        :param str region: (optional)
            Beaker cache region.
        :param dict kwargs:
            Keyworded arguments.
        :return:
            The value associated with the key or ``None``.
        """
        cache = self._get_cache(namespace, region, **kwargs)
        try:
            return cache.get(key)
        except KeyError:
            return None

    # -------------------------------------------------------------------------
    def clear(self, key=None, namespace='', region=None):
        """Clear a key/value or the entire namespace ``namespace``.

        :param str key: (optional)
            Key of value to remove.
        :param str namespace: (default='')
            Cache namespace.
        :param str region: (optional)
            Beaker cache region.
        """
        cache = self._get_cache(namespace, region)
        if key is not None:
            cache.remove_value(key=key)
        else:
            cache.clear()
            del self._caches[namespace]

    # -------------------------------------------------------------------------
    def initialize(self, namespace='', region=None):
        """Clear a key/value or the entire namespace ``namespace`` and set
        the time of creation.

        :param str namespace: (default='')
            Cache namespace.
        :param str region: (optional)
            Beaker cache region.
        """
        cache = self._get_cache(namespace, region)
        cache.clear()
        cache.put(CREATED_KEY, time())

    # -------------------------------------------------------------------------
    def _get_cache(self, namespace, region, **kwargs):
        """Return a Cache object.

        :param str namespace:
            Cache namespace.
        :param str region:
            Name of region.
        :param dict kwargs:
            Keyworded arguments.
        :rtype: beaker.cache.Cache
        """
        if namespace in self._caches:
            return self._caches[namespace]

        if region is None or region not in self.regions:
            if 'expire' not in kwargs:
                kwargs['expire'] = CACHE_DEFAULT_TTL
            self._caches[namespace] = self._cache_manager.get_cache(
                namespace, **kwargs)
        else:
            if not self.regions[region]['expire']:
                self.regions[region]['expire'] = CACHE_DEFAULT_TTL
            region_kwargs = dict(self.regions[region])
            region_kwargs.update(kwargs)
            self._caches[namespace] = self._cache_manager.get_cache(
                namespace, **region_kwargs)

        return self._caches[namespace]


# =============================================================================
def cache_global_item(namespace_prefix, region=None, access_function=None):
    """A decorator to retrieve in the global cache the result of a creation
    method whose the two first arguments are ``request`` and ``item_id``.
    The calling class must have a ``_<item>s`` attribute where `<item>` is the
    class name in lower case. If ``access_function`` is not ``None``, the
    access rights are checked.

    :param str namespace_prefix:
        Prefix of the cache namespace.
    :param tuple region: (optional)
        Name of region.
    :param access_function: (optional)
        A function to retrieve the access tuple.
    """
    def _decorated(method):
        """Decoration of the method `method`."""

        def _wrapper(class_, request, item_id, *args, **kwargs):
            """Use of global cache."""
            create_method = method.__func__ \
                if isinstance(method, classmethod) else method
            if not hasattr(class_, '_{0}s'.format(create_method.__name__)):
                raise Exception('Class must have a "_{0}s" attribute!'.format(
                    create_method.__name__))

            items = getattr(class_, '_{0}s'.format(create_method.__name__))
            namespace = cache_namespace(namespace_prefix, item_id)
            cache_time = request.registry['cache_global'].get(
                CREATED_KEY, namespace, region)
            if cache_time is None:
                cache_time = time()
                request.registry['cache_global'].set(
                    CREATED_KEY, cache_time, namespace, region)

            # Get the object
            if item_id not in items or items[item_id].created < cache_time:
                if item_id in items:
                    del items[item_id]
                item = create_method(class_, request, item_id, *args, **kwargs)
                if item is None:
                    return None
                items[item_id] = item

            # Check access rights
            if access_function is None:
                return items[item_id]
            get_access = access_function.__func__ \
                if isinstance(access_function, classmethod) \
                else access_function
            access = get_access(class_, request, items[item_id])
            return items[item_id] if access[0] else None

        return _wrapper

    return _decorated


# =============================================================================
def cache_global_value(key, namespace_prefix, region=None):
    """A decorator to retrieve in the global cache the result of a computing
    method whose the first argument is ``request``. The class of the method
    must have an ``uid`` attribute.

    :param str key:
        Key of the value to retrieve.
    :param str namespace_prefix:
        Prefix of the cache namespace.
    :param str region: (optional)
        Name of region.
    """
    def _decorated(method):
        """Decoration of the method `method`."""

        def _wrapper(class_, request, *args, **kwargs):
            """Use of global cache."""
            if not hasattr(class_, 'uid'):
                raise Exception('Class must have a "uid" attribute!')

            namespace = cache_namespace(namespace_prefix, class_.uid)
            value = request.registry['cache_global'].get(
                key, namespace, region)
            if value is not None and not kwargs.get('refresh'):
                return value

            computing_method = method.__func__ \
                if isinstance(method, classmethod) else method
            value = computing_method(class_, request, *args, **kwargs)
            request.registry['cache_global'].set(key, value, namespace, region)
            return value

        return _wrapper

    return _decorated


# =============================================================================
def cache_namespace(namespace_prefix, item_id):
    """Return a namespace for the item ``item_id``.

    :param str namespace:
        Cache namespace.
    :param str item_id:
        ID of the item.
    :rtype: str
    """
    return '{0}-{1}'.format(namespace_prefix, item_id)



# pylint: disable = import-outside-toplevel
"""Tests of ``subscribers`` classes and functions."""

from unittest import TestCase

from pyramid import testing
from pyramid.i18n import TranslationStringFactory

_ = TranslationStringFactory('chrysalio')


# =============================================================================
class SubscribersTestCase(TestCase):
    """Base class for testing subscribers."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = testing.setUp(settings={
            'testing': 'true',
            'theme.roots': 'chrysalio:Themes',
            'theme.patterns': 'Default, Alternative*'})

    # -------------------------------------------------------------------------
    def tearDown(self):
        """Undo the effects ``pyramid.testing.setUp()``."""
        testing.tearDown()


# =============================================================================
class USubscribersIncludeme(SubscribersTestCase):
    """Unit test class for :func:`subscribers.includeme`."""

    # -------------------------------------------------------------------------
    def setUp(self):
        """Set up test application."""
        self.configurator = testing.setUp()

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:routes.includeme]"""
        self.configurator.include('..subscribers')
        self.assertIsNotNone(
            self.configurator.registry.introspector.get_category(
                'subscribers'))


# =============================================================================
class USubscribersBeforeRender(SubscribersTestCase):
    """Unit test class for :func:`subscribers.before_render`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:routes.before_render]"""
        from ..includes.themes import load_themes
        from ..subscribers import before_render

        self.configurator.include('pyramid_chameleon')
        self.configurator.registry['settings'] = {
            'title': 'Test Chrysalio', 'theme': ''}
        load_themes(self.configurator)
        request = testing.DummyRequest()
        request.session['theme'] = 'Default'
        event = {
            'request': request, 'renderer_name': 'chrysalio:Templates/home.pt'}
        before_render(event)
        self.assertIn('layout', event)
        self.assertIn('theme', event)
        self.assertIn('theme_has', event)
        self.assertIn('_', event)
        self.assertIn('title', event)

        msg = event['_'](_('Foo'))
        self.assertEqual(msg, 'Foo')



# pylint: disable = import-outside-toplevel
"""Tests of ``lib.log`` functions."""

from logging import getLogger
from os.path import join, dirname, exists
from tempfile import mkdtemp
from shutil import rmtree
from configparser import ConfigParser
from unittest import TestCase

from testfixtures import LogCapture

from pyramid.testing import DummyRequest

from . import TmpDirTestCase


# =============================================================================
class ULibLogSetupLogging(TestCase):
    """Unit test class for :func:`lib.log.setup_logging`."""

    # -------------------------------------------------------------------------
    def test_console(self):
        """[u:lib.log.setup_logging] console"""
        from ..lib.log import setup_logging

        setup_logging()
        log = getLogger()
        self.assertGreater(len(log.handlers), 0)
        # self.assertIsInstance(log.handlers[0], StreamHandler)

    # -------------------------------------------------------------------------
    def test_file(self):
        """[u:lib.log.setup_logging] file"""
        from ..lib.log import setup_logging

        log_dir = mkdtemp()
        try:
            setup_logging(log_file=join(log_dir, 'test.log'))
            log = getLogger()
            self.assertGreater(len(log.handlers), 0)
        finally:
            rmtree(log_dir)


# =============================================================================
class ULibLogLogActivitySetup(TmpDirTestCase):
    """Unit test class for :func:`lib.log.log_activity_setup`."""

    # -------------------------------------------------------------------------
    def test_it(self):
        """[u:lib.log.log_activity_setup]"""
        from . import TEST_DIR, TEST2_INI, TEST3_INI
        from ..lib.utils import tounicode
        from ..lib.log import log_activity_setup

        config = ConfigParser({'here': dirname(TEST2_INI)})
        config.read(tounicode(TEST2_INI), encoding='utf8')
        self.assertFalse(log_activity_setup(config))

        config = ConfigParser({'here': dirname(TEST3_INI)})
        config.read(tounicode(TEST3_INI), encoding='utf8')
        self.assertTrue(log_activity_setup(config))
        self.assertTrue(exists(TEST_DIR))


# =============================================================================
class ULibLogLogActivity(TestCase):
    """Unit test class for :func:`lib.log.log_info`."""

    # -------------------------------------------------------------------------
    @classmethod
    def test_it(cls):
        """[u:lib.log.log_info]"""
        from ..lib.log import LOG_ACTIVITY, log_info

        with LogCapture() as log_capture:
            request = DummyRequest()
            request.session['user'] = {'user_id': 1, 'login': 'test1'}
            request.registry['log_activity'] = LOG_ACTIVITY
            log_info(request, 'action', 'foo', 'bar')
            log_capture.check(('activity', 'INFO', '[test1] action foo bar'))


# =============================================================================
class ULibLogLogError(TestCase):
    """Unit test class for :func:`lib.log.log_error`."""

    # -------------------------------------------------------------------------
    @classmethod
    def test_it(cls):
        """[u:lib.log.log_error]"""
        from ..lib.log import LOG_ACTIVITY, log_error

        with LogCapture() as log_capture:
            request = DummyRequest()
            request.session['user'] = {'user_id': 1, 'login': 'test1'}
            request.registry['log_activity'] = LOG_ACTIVITY
            log_error(request, 'An error occured!')
            log_capture.check(
                ('activity', 'ERROR', '[test1] An error occured!'))


# =============================================================================
class ULibLogLogWarning(TestCase):
    """Unit test class for :func:`lib.log.log_warning`."""

    # -------------------------------------------------------------------------
    @classmethod
    def test_it(cls):
        """[u:lib.log.log_warning]"""
        from ..lib.log import LOG_ACTIVITY, log_warning

        with LogCapture() as log_capture:
            request = DummyRequest()
            request.session['user'] = {'user_id': 2, 'login': 'test2'}
            request.registry['log_activity'] = LOG_ACTIVITY
            log_warning(request, 'Warning!')
            log_capture.check(('activity', 'WARNING', '[test2] Warning!'))


# =============================================================================
class ULibLogLogDebug(TestCase):
    """Unit test class for :func:`lib.log.log_debug`."""

    # -------------------------------------------------------------------------
    @classmethod
    def test_it(cls):
        """[u:lib.log.log_warning]"""
        from ..lib.log import LOG_ACTIVITY, log_debug

        with LogCapture() as log_capture:
            request = DummyRequest()
            request.session['user'] = {'user_id': 3, 'login': 'test3'}
            request.registry['log_activity'] = LOG_ACTIVITY
            log_debug(request, 'Debug message')
            log_capture.check(('activity', 'DEBUG', '[test3] Debug message'))



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``scripts`` classes."""

from collections import namedtuple

from . import DBUnitTestCase


# =============================================================================
class UScriptsScript(DBUnitTestCase):
    """Unit test class for :class:`scripts.Script`."""

    # -------------------------------------------------------------------------
    def test_init(self):
        """[u:scripts.Script]"""
        from . import TEST_INI, TEST1_INI
        from ..scripts import Script

        args = namedtuple('ArgumentParser', 'conf_uri options lang')(
            conf_uri='foo.ini', options=None, lang=None)
        script = Script(args, False)
        self.assertIsNone(script.registry)

        args = namedtuple('ArgumentParser', 'conf_uri options lang')(
            conf_uri=TEST1_INI, options=None, lang=None)
        script = Script(args, False)
        self.assertIsNone(script.registry)

        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang drop_tables')(
                conf_uri=TEST_INI, options=None, lang=None, drop_tables=True)
        script = Script(args, False)
        self.assertIsNotNone(script.registry)

        script = Script(args, True)
        self.assertIsNotNone(script.registry)

    # -------------------------------------------------------------------------
    def test_init_bad_include(self):
        """[u:scripts.Script] bad included module"""
        from . import TEST5_INI
        from ..scripts import Script

        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang drop_tables')(
                conf_uri=TEST5_INI, options=None, lang=None, drop_tables=False)
        self.assertRaises(SystemExit, Script, args, False)

    # -------------------------------------------------------------------------
    def test_init_no_includeme(self):
        """[u:scripts.Script] included module without includeme()"""
        from . import TEST6_INI
        from ..scripts import Script

        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang drop_tables')(
                conf_uri=TEST6_INI, options=None, lang=None, drop_tables=False)
        self.assertRaises(SystemExit, Script, args, True)

    # -------------------------------------------------------------------------
    def test_init_bad_dependency(self):
        """[u:scripts.Script] included module with unsatisfied dependency"""
        from . import TEST7_INI
        from ..scripts import Script

        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang drop_tables')(
                conf_uri=TEST7_INI, options=None, lang=None, drop_tables=False)
        self.assertRaises(SystemExit, Script, args, True)

    # -------------------------------------------------------------------------
    def test_init_conflict(self):
        """[u:scripts.Script] included conflicting modules"""
        from . import TEST8_INI
        from ..scripts import Script

        args = namedtuple(
            'ArgumentParser', 'conf_uri options lang drop_tables')(
                conf_uri=TEST8_INI, options=None, lang=None, drop_tables=False)
        self.assertRaises(SystemExit, Script, args, True)

    # -------------------------------------------------------------------------
    def test_argument_parser(self):
        """[u:scripts.Script.argument_parser]"""
        from argparse import ArgumentParser
        from ..scripts import Script

        parser = Script.argument_parser('éÔ')
        self.assertIsInstance(parser, ArgumentParser)
        self.assertEqual(parser.description, 'éÔ')

    # -------------------------------------------------------------------------
    def test_arguments(self):
        """[u:scripts.Script.argument]"""
        from . import TEST_INI
        from ..scripts import Script

        parser = Script.argument_parser('My description')
        self.assertRaises(SystemExit, Script.arguments, parser)

        args = namedtuple('ArgumentParser', 'conf_uri')(conf_uri='foo.ini')
        self.assertIsNone(Script.arguments(parser, args))

        args = namedtuple('ArgumentParser', 'conf_uri')(conf_uri=TEST_INI)
        args = Script.arguments(parser, args)
        self.assertIn('conf_uri', args)
        self.assertEqual(args.conf_uri, TEST_INI)
        self.assertIn('log_level', args)
        self.assertEqual(args.log_level, 'INFO')



# -*- coding: utf-8 -*-
# pylint: disable = import-outside-toplevel
"""Tests of ``lib.ftp`` classes."""

from os.path import join, exists, basename
from ftplib import FTP, all_errors

from . import TmpDirTestCase


# =============================================================================
class ULibFtpFtp(TmpDirTestCase):
    """Unit test class for :class:`lib.ftp.Ftp`."""

    # -------------------------------------------------------------------------
    def _make_one(self):
        """Return a connected Ftp object.

        :rtype: :class:`.lib.ftp.Ftp` or ``None``
        """
        from chrysalio.lib.utils import decrypt
        from . import FTP_HOST, FTP_USER, FTP_PASSWORD
        from ..lib.ftp import Ftp

        values = {
            'ftp_host': FTP_HOST, 'ftp_user': FTP_USER,
            'ftp_password': decrypt(FTP_PASSWORD, 'test'), 'ftp_pasv': True}
        ftp = Ftp(self.log_error)
        return ftp if ftp.connect(values) else None

    # -------------------------------------------------------------------------
    def log_error(self, text):
        """Dummy method for ``log_error``.

        :param str text:
            Error.
        """

    # -------------------------------------------------------------------------
    def test_connect(self):
        """[u:lib.ftp.Ftp.connect]"""
        from chrysalio.lib.utils import decrypt
        from . import FTP_HOST, FTP_USER, FTP_PASSWORD
        from ..lib.ftp import Ftp

        ftp = Ftp(self.log_error)

        # Missing values
        self.assertFalse(ftp.connect({}))

        # Invalid host
        values = {
            'ftp_host': 'ftp.chrysal2.io', 'ftp_user': FTP_USER,
            'ftp_password': 'foo', 'ftp_pasv': True, 'ftp_path': 'Foo'}
        self.assertFalse(ftp.connect(values))

        # Invalid password
        values['ftp_host'] = FTP_HOST
        self.assertFalse(ftp.connect(values))

        # Not existing directory
        values['ftp_password'] = decrypt(FTP_PASSWORD, 'test')
        self.assertFalse(ftp.connect(values))

        # Connection
        del values['ftp_path']
        ftp.connection = FTP()
        try:
            ftp.connection.connect(
                values.get('ftp_host'), values.get('ftp_port', 21))
        except all_errors:
            return
        self.assertTrue(
            ftp.connect(values), msg='Put somes files in {0}@{1}'.format(
                FTP_USER, FTP_HOST))
        ftp.quit()

    # -------------------------------------------------------------------------
    def test_cwd(self):
        """[u:lib.ftp.Ftp.cwd]"""
        ftp = self._make_one()
        if not ftp:
            return

        # Not existing directory
        self.assertFalse(ftp.cwd('Foo'))

        # Existing directoy
        ftp.mkdir('Bar')
        self.assertTrue(ftp.cwd('Bar'))
        ftp.rmtree('Bar')
        ftp.quit()

    # -------------------------------------------------------------------------
    def test_mkdir(self):
        """[u:lib.ftp.Ftp.mkdir]"""
        ftp = self._make_one()
        if not ftp:
            return

        self.assertTrue(ftp.mkdir('Foo'))
        ftp.rmtree('Foo')
        ftp.quit()

    # -------------------------------------------------------------------------
    def test_list_directory(self):
        """[u:lib.ftp.Ftp.list_directory]"""

        ftp = self._make_one()
        if not ftp:
            return
        ftp.mkdir('Bar')

        dirs, files = ftp.list_directory()
        self.assertIsInstance(dirs, dict)
        self.assertIsInstance(files, dict)
        self.assertTrue(dirs)
        ftp.rmtree('Bar')
        ftp.quit()

    # -------------------------------------------------------------------------
    def test_rmtree(self):
        """[u:lib.ftp.Ftp.rmtree]"""
        from . import TEST_INI

        ftp = self._make_one()
        if not ftp:
            return

        # Not existing directory
        self.assertFalse(ftp.rmtree('Foo'))

        # Existing directory
        ftp.mkdir('Bar')
        ftp.cwd('Bar')
        ftp.mkdir('Baz')
        with open(TEST_INI, 'rb') as hdl:
            ftp.connection.storbinary('STOR test.ini', hdl)
        ftp.cwd('..')
        self.assertTrue(ftp.rmtree('Bar'))
        ftp.quit()

    # -------------------------------------------------------------------------
    def test_download(self):
        """[u:lib.ftp.Ftp.download]"""
        from ..lib.utils import tostr
        from . import TEST_INI, TEST_DIR

        ftp = self._make_one()
        if not ftp:
            return
        ftp.rmtree('Baz')
        ftp.mkdir('Baz')
        ftp.connection.cwd('Baz')
        destination = join(TEST_DIR, 'Download')

        # Empty
        self.assertTrue(ftp.download(destination))
        self.assertFalse(exists(destination))

        # Not empty
        ftp.mkdir('Foo')
        ftp.mkdir('.svn')
        ftp.upload(TEST_INI)
        with open(TEST_INI, 'rb') as hdl:
            ftp.connection.storbinary(tostr('STOR écrit.ini.part'), hdl)
        with open(TEST_INI, 'rb') as hdl:
            ftp.connection.storbinary('STOR Thumbs.db', hdl)

        self.assertFalse(ftp.download(destination))
        self.assertTrue(exists(destination))
        ftp.delete(tostr('écrit.ini.part'))
        self.assertTrue(ftp.download(destination))
        self.assertTrue(exists(join(destination, basename(TEST_INI))))
        self.assertFalse(exists(join(destination, 'Thumbs.db')))
        self.assertFalse(exists(join(destination, '.svn')))

        ftp.connection.cwd('..')
        ftp.rmtree('Baz')
        ftp.quit()

        self.assertTrue(exists(join(destination, 'test.ini')))
        self.assertTrue(exists(join(destination, 'Foo')))

    # -------------------------------------------------------------------------
    def test_upload(self):
        """[u:lib.ftp.Ftp.uplaod]"""
        from . import TEST_INI

        ftp = self._make_one()
        if not ftp:
            return
        ftp.upload(TEST_INI)
        files = ftp.list_directory()[1]
        self.assertIn('test.ini', files)
        ftp.delete('test.ini')
        ftp.quit()

    # -------------------------------------------------------------------------
    def test_quit(self):
        """[u:lib.ftp.Ftp.quit]"""
        ftp = self._make_one()
        if not ftp:
            return

        self.assertIsNotNone(ftp.connection)
        ftp.quit()
        self.assertIsNone(ftp.connection)



"""Main menu management."""

from pyramid.config import Configurator
from pyramid.events import BeforeRender
from pyramid.threadlocal import get_current_request

from .lib.i18n import _
from .lib.menu import Menu


# Menu entry = (icon, label, permission, route, (subentry, ...))
MENU_HOME = ('{theme}/images/menu_home.png', _('Home'), None, 'home', None)
MENU_TEST = (
    '{theme}/custom/menu_test1.png', _('Test'), None, None,
    (('{theme}/custom/menu_test2.png', _('Users'), None, 'user_index',
      ((None, _('Administrator'), None, ('user_view', {'user_id': 1}), None),
       (None, _('User 1'), None, ('user_view', {'user_id': 3}), None))),
     ('{theme}/custom/menu_test3.png', _('Profiles'), None, 'profile_index',
      None)))
MENU_ADMIN = (
    '{theme}/images/menu_admin.png', _('Administration'), None, None,
    [(None, _('General Settings'), 'settings-view', 'settings_view', None),
     (None, _('Groups'), 'group-view', 'group_index', None),
     (None, _('Users'), 'user-view', 'user_index', None),
     (None, _('Profiles'), 'profile-view', 'profile_index', None),
     (None, _('Backup'), 'backup-create', 'backup', None)])


# =============================================================================
def includeme(configurator):
    """Function to include menu functionality.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    if isinstance(configurator, Configurator):
        configurator.registry['menu'] = [MENU_HOME, MENU_ADMIN]
        configurator.add_subscriber(before_render, BeforeRender)


# =============================================================================
def before_render(event):
    """A subscriber for :class:`pyramid.events.BeforeRender` events to add a
    main menu.

    :type  event: pyramid.events.BeforeRender
    :param event:
        Current event.
    """
    request = event.get('request') or get_current_request()
    event['menu'] = Menu(request, 'menu', request.registry.get('menu', ''))



"""Configuration view callables."""

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError
from lxml import etree

from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound
from pyramid.response import Response

from ....lib.i18n import _
from ....lib.log import log_info
from ....lib.utils import make_id
from ....lib.form import get_action, Form
from ....lib.xml import validate_xml, relaxng4validation
from ....views import BaseView
from ....models.populate import web2db
from ....models.dbprofile import DBProfile
from ..models.populate import xml2db
from ..models.dbldap import DBLdap, DBLdapProfile
from ..relaxng import RELAXNG_CIOLDAP


# =============================================================================
class CioLDAPView(BaseView):
    """Class to configure CioLDAP module.

    :type  request: pyramid.request.Request
    :param request:
        Current request.
    """

    # -------------------------------------------------------------------------
    @view_config(
        route_name='cioldap_view',
        renderer='chrysalio:modules/cioldap/Templates/cioldap_view.pt',
        permission='modules-view')
    @view_config(route_name='cioldap_view', renderer='json', xhr=True)
    def view(self):
        """Display CioLDAP configuration."""
        i_editor = self._request.has_permission('cioldap-edit')
        if self._request.is_xhr:
            if i_editor:
                self._web2db()
            return {}

        # Edition mode
        dbldap = self._request.dbsession.query(DBLdap).first()
        if dbldap is None and self._request.has_permission('modules-edit'):
            return HTTPFound(self._request.route_path('cioldap_edit'))

        # Initialization
        profile_labels = {
            k.profile_id: k.label(self._request)
            for k in self._request.dbsession.query(DBProfile)}

        # Action
        action = get_action(self._request)[0]
        if action == 'exp!' and dbldap is not None:
            action = self._db2web(dbldap)
            if action:
                return action
        elif action == 'imp!' and i_editor:
            self._web2db()
            dbldap = self._request.dbsession.query(DBLdap).first()

        self._request.breadcrumbs(
            _('CioLDAP Module Configuration'),
            replace=self._request.route_path('cioldap_edit'))

        return {
            'form': Form(self._request), 'dbldap': dbldap,
            'i_editor': i_editor, 'profile_labels': profile_labels,
            'action': action,
            'download_max_size': self._request.registry['settings'][
                'download-max-size']}

    # -------------------------------------------------------------------------
    @view_config(
        route_name='cioldap_edit',
        renderer='chrysalio:modules/cioldap/Templates/cioldap_edit.pt',
        permission='cioldap-edit')
    def edit(self):
        """Create or edit CioLDAP configuration."""
        dbldap = self._request.dbsession.query(DBLdap).first()
        creation = dbldap is None
        profile_labels = {
            k.profile_id: k.label(self._request)
            for k in self._request.dbsession.query(DBProfile)}

        # Form and action
        form = Form(
            self._request, *DBLdap.settings_schema(profile_labels, dbldap),
            obj=dbldap)
        action = get_action(self._request)[0]
        if action == 'imp!':
            web2db(
                self._request, xml2db, 'ldap',
                relaxngs={
                    '{{{0}}}{1}'.format(
                        RELAXNG_CIOLDAP['namespace'],
                        RELAXNG_CIOLDAP['root']): RELAXNG_CIOLDAP['file']})
            if not self._request.session.peek_flash('alert'):
                log_info(self._request, 'ldap_import')
                return HTTPFound(self._request.route_path('cioldap_view'))
        elif action == 'sav!' and form.validate():
            dbldap = self._save(dbldap, profile_labels, form.values)
            if dbldap is not None:
                self._request.registry['authorities'][
                    'ldap'].reset_configuration()
                log_info(
                    self._request, creation and 'ldap_edit' or 'ldap_create')
                return HTTPFound(self._request.route_path('cioldap_view'))
        if form.has_error():
            self._request.session.flash(_('Correct errors.'), 'alert')

        self._request.breadcrumbs(
            _('CioLDAP Module Edition'), replace=self._request.route_path(
                'cioldap_view'))

        return {
            'form': form, 'dbldap': dbldap or DBLdap,
            'profile_labels': profile_labels,
            'creation': dbldap is None, 'action': action}

    # -------------------------------------------------------------------------
    def _save(self, dbldap, profiles, values):
        """Save a CioLDAP configuration.

        :type  dbldap: .models.dbldap.DBLdap
        :param dbldap:
            Ldap to save.
        :param dict profiles:
            A dictionary such as ``{profile_id: label,...}``.
        :param dict values:
            Form values.
        :rtype: :class:`~.models.dbldap.Ldap` instance or ``None``
        """
        creation = dbldap is None
        dbldap = dbldap or DBLdap()

        # Update ldap
        record = {k: values[k] for k in values if hasattr(DBLdap, k)}
        record['root_password'] = \
            values['root_password1'] or \
            (not creation and dbldap.root_password) or None
        error = dbldap.record_format(record)
        if error:  # pragma: nocover
            self._request.session.flash(error, 'alert')
            return None
        for field in record:
            setattr(dbldap, field, record[field])

        # Save
        if creation:
            try:
                self._request.dbsession.add(dbldap)
                self._request.dbsession.flush()
            except (IntegrityError, FlushError):  # pragma: nocover
                self._request.session.flash(
                    _('This CioLDAP configuration already exists.'), 'alert')
                return None

        # Update profiles
        ldap_profiles = {k.profile_id: k for k in dbldap.user_profiles}
        for profile_id in profiles:
            value = values['pfl:{0}'.format(profile_id)]
            if value and profile_id not in ldap_profiles:
                self._request.dbsession.add(DBLdapProfile(
                    ldap_id=dbldap.ldap_id, profile_id=profile_id))
            elif not value and profile_id in ldap_profiles:
                self._request.dbsession.query(DBLdapProfile).filter_by(
                    ldap_id=dbldap.ldap_id, profile_id=profile_id).delete()

        return dbldap

    # -------------------------------------------------------------------------
    def _db2web(self, dbldap):
        """Export CioLDAP configuration as an XML file embedded in a Pyramid
        response.

        :type  dbldap: :class:`.modules.cioldap.models.dbldap.DBLdap`
        :param dbldap:
            Current SqlAlchemy LDPA object.
        :rtype: :class:`pyramid.response.Response` or ``''``
        """
        # Create the XML file
        root = '{{{0}}}{1}'.format(
            RELAXNG_CIOLDAP['namespace'], RELAXNG_CIOLDAP['root'])
        root_elt = etree.Element(
            root, version=RELAXNG_CIOLDAP['version'],
            nsmap={None: RELAXNG_CIOLDAP['namespace']})
        root_elt.append(dbldap.db2xml())
        root_elt = etree.XML(etree.tostring(root_elt, encoding='utf-8'))
        error = validate_xml(
            etree.ElementTree(root_elt), relaxng4validation(RELAXNG_CIOLDAP),
            True)
        if error:  # pragma: nocover
            return self._request.localizer.translate(error)

        # Convert it into a response
        filename = '{0}.{1}.xml'.format(
            make_id(self._request.registry['settings']['title'], 'token'),
            DBLdap.suffix)
        response = Response(
            body=etree.tostring(
                root_elt, pretty_print=True, encoding='utf-8',
                xml_declaration=True), content_type='application/xml')
        response.headerlist.append((
            'Content-Disposition', 'attachment; filename="{0}"'.format(
                filename)))
        log_info(self._request, 'cioldap_export')
        return response

    # -------------------------------------------------------------------------
    def _web2db(self):
        """Import a CioLDAP configuration."""
        web2db(
            self._request, xml2db, 'ldap', relaxngs={
                '{{{0}}}{1}'.format(
                    RELAXNG_CIOLDAP['namespace'],
                    RELAXNG_CIOLDAP['root']): RELAXNG_CIOLDAP['file']})
        log_info(self._request, 'cioldap_import')



"""Clipboard management."""

from pyramid.config import Configurator

from ..lib.i18n import _
from ..lib.panel import PANEL_ITEM_PREFIX, Panel
from ..models.dbsettings import SETTINGS_DEFAULTS


CLIPBOARD_LABEL_MAX_LENGTH = 96
CLIPBOARD_LABEL_MAX_ITEMS = 8


# =============================================================================
def includeme(configurator):
    """Function to include themes.

    :type  configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    if isinstance(configurator, Configurator):
        if 'panels' not in configurator.registry:
            configurator.registry['panels'] = {}
        if 'clipboard' not in configurator.registry['panels']:
            Clipboard.register(configurator.registry, Clipboard)


# =============================================================================
class Clipboard(Panel):
    """Class to manage clipboard and clipboard panel.

    See: :class:`.lib.panel.Panel`
    """

    uid = 'clipboard'
    label = _('Clipboard')
    icon = '/images/panel_clipboard.png'
    template = 'chrysalio:Templates/panel_clipboard.pt'
    need_form = True

    # -------------------------------------------------------------------------
    @classmethod
    def is_empty(cls, request):
        """Return ``True`` if the clipboard is empty.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: bool
        """
        return bool(not request.session.get('clipboard'))

    # -------------------------------------------------------------------------
    @classmethod
    def entries(cls, request):
        """Return the list of entries.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :rtype: list
        """
        return request.session['clipboard'] \
            if 'clipboard' in request.session else []

    # -------------------------------------------------------------------------
    @classmethod
    def push(cls, request, domain, cut, data, preview):
        """Push a new entry in the clipboard buffer.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param str domain:
            Domain of the entry (ex.: ``'warehouse/file'``).
        :param bool cut:
            ``True`` if it is a `cut` operation.
        :param data:
            Content of this entry. It  depends on the domain.
        :type  preview: webhelpers2.html.literal
        :param preview:
            Representation of the entry in the clipboard.
        """
        # Ajust size
        if 'clipboard' not in request.session:
            request.session['clipboard'] = []
        request.session['clipboard'] = request.session['clipboard'][
            :request.registry['settings'].get(
                'clipboard-size', SETTINGS_DEFAULTS['clipboard-size']) - 1]

        # Push the new entry in the buffer
        request.session['clipboard'].insert(0, (domain, cut, data, preview))

    # -------------------------------------------------------------------------
    @classmethod
    def selection(cls, request, domains):
        """Return entries selected in the clipboard and compatible with almost
        one domain of ``domains``.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param list domains:
            List of accepted formats of the entry.
        :rtype: list
        """
        if 'clipboard' not in request.session \
           or not request.session['clipboard']:
            return ()

        # Selected entries
        entries = []
        kept_nums = []
        requested_nums = cls._requested_numbers(request, domains)
        for num in requested_nums:
            entries.append(request.session['clipboard'][num])
            if not request.session['clipboard'][num][1]:
                kept_nums.append(num)

        # Reorganization of the buffer
        my_buffer = []
        for num in kept_nums:
            my_buffer.append(request.session['clipboard'][num])
        for num in range(len(request.session['clipboard'])):
            if num not in requested_nums:
                my_buffer.append(request.session['clipboard'][num])
        request.session['clipboard'] = my_buffer

        return entries

    # -------------------------------------------------------------------------
    @classmethod
    def _requested_numbers(cls, request, domains):
        """Return a list of numbers of entry requested and compatible with
        ``domains``.

        :type  request: pyramid.request.Request
        :param request:
            Current request.
        :param list domains:
            List of accepted formats of the entry.
        :rtype: list
        """
        selected = []
        size = len(request.session['clipboard'])

        # Selected in the clipboard panel
        for num in request.params:
            if num.startswith(PANEL_ITEM_PREFIX):
                num = int(num[len(PANEL_ITEM_PREFIX):])
                if num < size and \
                   request.session['clipboard'][num][0] in domains:
                    selected.append(num)
        if selected:
            return selected

        # The first compatible entry
        for num in range(size):
            if request.session['clipboard'][num][0] in domains:
                selected.append(num)
                break
        return selected



"""Database main objects, functions and constants."""

from sys import exit as sys_exit
from sqlalchemy import engine_from_config
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.schema import MetaData
from sqlalchemy.exc import ProgrammingError, OperationalError
import zope.sqlalchemy

from ..lib.i18n import _, translate


DB_METADATA = MetaData(naming_convention={
    'ix': 'ix_%(column_0_label)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(constraint_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'pk': 'pk_%(table_name)s'
})

ID_LEN = 32
NAME_LEN = 64
EMAIL_LEN = 64
LABEL_LEN = 128
DESCRIPTION_LEN = 255
VALUE_LEN = 255
MODULE_LEN = 128

DBDeclarativeClass = declarative_base(metadata=DB_METADATA)


# =============================================================================
def includeme(configurator):
    """Initialize the model for a Chrysalio application.

    :type configurator: pyramid.config.Configurator
    :param configurator:
        Object used to do configuration declaration within the application.
    """
    settings = configurator.get_settings()
    settings['tm.manager_hook'] = 'pyramid_tm.explicit_manager'

    # Use pyramid_tm to hook the transaction lifecycle to the request
    configurator.include('pyramid_tm')

    # Use pyramid_retry to retry a request when transient exceptions occur
    configurator.include('pyramid_retry')

    # Get database engine
    try:
        dbengine = get_dbengine(settings)
    except KeyError:
        sys_exit(translate(_('*** Database is not defined.')))
    DB_METADATA.bind = dbengine

    # Make request.dbsession available for use in Pyramid
    dbsession_factory = get_dbsession_factory(dbengine)
    configurator.registry['dbsession_factory'] = dbsession_factory
    configurator.add_request_method(
        # r.tm is the transaction manager used by pyramid_tm
        lambda r: get_tm_dbsession(dbsession_factory, r.tm),
        'dbsession', reify=True)


# =============================================================================
def get_dbengine(settings, prefix='sqlalchemy.'):
    """Get SQLAlchemy engine.

    :type  settings: pyramid.registry.Registry.settings
    :param settings:
        Application settings.
    :param str prefix: (default='sqlalchemy')
        Prefix in settings for SQLAlchemy configuration.
    :rtype: sqlalchemy.engine.base.Engine
    """
    return engine_from_config(settings, prefix)


# =============================================================================
def get_dbsession_factory(dbengine):
    """Get SQLAlchemy session factory.

    :type  dbengine: sqlalchemy.engine.base.Engine
    :param dbengine:
        Database engine.
    :rtype: :func:`sqlalchemy.orm.session.sessionmaker`
    """
    factory = sessionmaker()
    factory.configure(bind=dbengine)
    return factory


# =============================================================================
def get_tm_dbsession(dbsession_factory, transaction_manager):
    """Get a ``sqlalchemy.orm.Session`` instance backed by a transaction.

    :type  dbsession_factory: sqlalchemy.orm.session.sessionmaker
    :param dbsession_factory:
        Function to create session.
    :type transaction_manager: transaction._manager.ThreadTransactionManager
    :param transaction_manager:
        Transaction manager.
    :rtype: sqlalchemy.orm.session.Session

    This function will hook the session to the transaction manager which
    will take care of committing any changes.

    - When using pyramid_tm it will automatically be committed or aborted
      depending on whether an exception is raised.

    - When using scripts you should wrap the session in a manager yourself.

      For example::

          import transaction

          dbsession_factory = get_dbsession_factory(DB_METADATA.bind)
          with transaction.manager:
              dbsession = get_tm_dbsession(
                  dbsession_factory, transaction.manager)
    """
    dbsession = dbsession_factory()
    zope.sqlalchemy.register(
        dbsession, transaction_manager=transaction_manager)
    return dbsession


# =============================================================================
def add_column(table_class, column):
    """Add a column to a database table.

    :type  table_class: DBDeclarativeClass
    :param table_class:
        SQLAlchemy object of the table to proceed.
    :type  column: sqlalchemy.schema.Column
    :param column:
        Column to create.
    """
    table_class.__table__.append_column(column)

    dialect = DB_METADATA.bind.dialect
    table_name = str(column.compile(dialect=dialect))
    table_name = table_name.partition('.')[0] if '.' in table_name \
        else table_class.__table__.name
    try:
        DB_METADATA.bind.execute(
            'ALTER TABLE {table} ADD COLUMN {column} {type}'.format(
                table=table_name, column=column.name,
                type=column.type.compile(dialect=dialect)))
    except (ProgrammingError, OperationalError):  # pragma: nocover
        pass

    setattr(table_class, column.name, column)
